import sys
import time
import scipy
import pandas as pd
import scanpy as sc
import numpy as np
import sklearn
from sklearn import preprocessing
import scipy.sparse as sp
from anndata import AnnData

from tqdm import tqdm
from typing import Optional


# Chromatin structure data related
def aggregate_peaks_10x(adata_atac, peak_annot_file, linkage_file, peak_dist=2000, min_corr=0.5, gene_body=False, return_dict=False, verbose=False):
    """Peak to gene aggregation.
    This function aggregates promoter and enhancer peaks to genes based on the 10X linkage file.
    Adapted from the MultiVelo.
    Parameters
    ----------
    adata_atac: :class:`~anndata.AnnData`
        ATAC anndata object which stores raw peak counts.
    peak_annot_file: `str`
        Peak annotation file from 10X CellRanger ARC.
    linkage_file: `str`
        Peak-gene linkage file from 10X CellRanger ARC. This file stores highly correlated peak-peak
        and peak-gene pair information.
    peak_dist: `int` (default: 2000)
        Maximum distance for peaks to be included for a gene.
    min_corr: `float` (default: 0.5)
        Minimum correlation for a peak to be considered as enhancer.
    gene_body: `bool` (default: `False`)
        Whether to add gene body peaks to the associated promoters.
    return_dict: `bool` (default: `False`)
        Whether to return promoter and enhancer dictionaries.
    verbose: `bool` (default: `False`)
        Whether to print number of genes with promoter peaks.
    Returns
    -------
    A new ATAC anndata object which stores gene aggreagted peak counts.
    Additionally, if `return_dict==True`:
        A dictionary which stores genes and promoter peaks.
        And a dictionary which stores genes and enhancer peaks.
    """
    promoter_dict = {}
    distal_dict = {}
    gene_body_dict = {}
    corr_dict = {}

    # read annotations
    with open(peak_annot_file) as f:
        header = next(f)
        tmp = header.split('\t')
        if len(tmp) == 4:
            cellranger_version = 1
        elif len(tmp) == 6:
            cellranger_version = 2
        else:
            raise ValueError('Peak annotation file should contain 4 columns (CellRanger ARC 1.0.0) or 5 columns (CellRanger ARC 2.0.0)')
        if verbose:
            print(f'CellRanger ARC identified as {cellranger_version}.0.0')
        if cellranger_version == 1:
            for line in f:
                tmp = line.rstrip().split('\t')
                tmp1 = tmp[0].split('_')
                peak = f'{tmp1[0]}:{tmp1[1]}-{tmp1[2]}'
                if tmp[1] != '':
                    genes = tmp[1].split(';')
                    dists = tmp[2].split(';')
                    types = tmp[3].split(';')
                    for i,gene in enumerate(genes):
                        dist = dists[i]
                        annot = types[i]
                        if annot == 'promoter':
                            if gene not in promoter_dict:
                                promoter_dict[gene] = [peak]
                            else:
                                promoter_dict[gene].append(peak)
                        elif annot == 'distal':
                            if dist == '0':
                                if gene not in gene_body_dict:
                                    gene_body_dict[gene] = [peak]
                                else:
                                    gene_body_dict[gene].append(peak)
                            else:
                                if gene not in distal_dict:
                                    distal_dict[gene] = [peak]
                                else:
                                    distal_dict[gene].append(peak)
        else:
            for line in f:
                tmp = line.rstrip().split('\t')
                peak = f'{tmp[0]}:{tmp[1]}-{tmp[2]}'
                gene = tmp[3]
                dist = tmp[4]
                annot = tmp[5]
                if annot == 'promoter':
                    if gene not in promoter_dict:
                        promoter_dict[gene] = [peak]
                    else:
                        promoter_dict[gene].append(peak)
                elif annot == 'distal':
                    if dist == '0':
                        if gene not in gene_body_dict:
                            gene_body_dict[gene] = [peak]
                        else:
                            gene_body_dict[gene].append(peak)
                    else:
                        if gene not in distal_dict:
                            distal_dict[gene] = [peak]
                        else:
                            distal_dict[gene].append(peak)

    # read linkages
    with open(linkage_file) as f:
        for line in f:
            tmp = line.rstrip().split('\t')
            if tmp[12] == "peak-peak":
                peak1 = f'{tmp[0]}:{tmp[1]}-{tmp[2]}'
                peak2 = f'{tmp[3]}:{tmp[4]}-{tmp[5]}'
                tmp2 = tmp[6].split('><')[0][1:].split(';')
                tmp3 = tmp[6].split('><')[1][:-1].split(';')
                corr = float(tmp[7])
                for t2 in tmp2:
                    gene1 = t2.split('_')
                    for t3 in tmp3:
                        gene2 = t3.split('_')
                        # one of the peaks is in promoter, peaks belong to the same gene or are close in distance
                        if ((gene1[1] == "promoter") != (gene2[1] == "promoter")) and ((gene1[0] == gene2[0]) or (float(tmp[11]) < peak_dist)):
                            if gene1[1] == "promoter":
                                gene = gene1[0]
                            else:
                                gene = gene2[0]
                            if gene in corr_dict:
                                # peak 1 is in promoter, peak 2 is not in gene body -> peak 2 is added to gene 1
                                if peak2 not in corr_dict[gene] and gene1[1] == "promoter" and (gene2[0] not in gene_body_dict or peak2 not in gene_body_dict[gene2[0]]):
                                    corr_dict[gene][0].append(peak2)
                                    corr_dict[gene][1].append(corr)
                                # peak 2 is in promoter, peak 1 is not in gene body -> peak 1 is added to gene 2
                                if peak1 not in corr_dict[gene] and gene2[1] == "promoter" and (gene1[0] not in gene_body_dict or peak1 not in gene_body_dict[gene1[0]]):
                                    corr_dict[gene][0].append(peak1)
                                    corr_dict[gene][1].append(corr)
                            else:
                                # peak 1 is in promoter, peak 2 is not in gene body -> peak 2 is added to gene 1
                                if gene1[1] == "promoter" and (gene2[0] not in gene_body_dict or peak2 not in gene_body_dict[gene2[0]]):
                                    corr_dict[gene] = [[peak2], [corr]]
                                # peak 2 is in promoter, peak 1 is not in gene body -> peak 1 is added to gene 2
                                if gene2[1] == "promoter" and (gene1[0] not in gene_body_dict or peak1 not in gene_body_dict[gene1[0]]):
                                    corr_dict[gene] = [[peak1], [corr]]
            elif tmp[12] == "peak-gene":
                peak1 = f'{tmp[0]}:{tmp[1]}-{tmp[2]}'
                tmp2 = tmp[6].split('><')[0][1:].split(';')
                gene2 = tmp[6].split('><')[1][:-1]
                corr = float(tmp[7])
                for t2 in tmp2:
                    gene1 = t2.split('_')
                    # peak 1 belongs to gene 2 or are close in distance -> peak 1 is added to gene 2
                    if ((gene1[0] == gene2) or (float(tmp[11]) < peak_dist)):
                        gene = gene1[0]
                        if gene in corr_dict:
                            if peak1 not in corr_dict[gene] and gene1[1] != "promoter" and (gene1[0] not in gene_body_dict or peak1 not in gene_body_dict[gene1[0]]):
                                corr_dict[gene][0].append(peak1)
                                corr_dict[gene][1].append(corr)
                        else:
                            if gene1[1] != "promoter" and (gene1[0] not in gene_body_dict or peak1 not in gene_body_dict[gene1[0]]):
                                corr_dict[gene] = [[peak1], [corr]]
            elif tmp[12] == "gene-peak":
                peak2 = f'{tmp[3]}:{tmp[4]}-{tmp[5]}'
                gene1 = tmp[6].split('><')[0][1:]
                tmp3 = tmp[6].split('><')[1][:-1].split(';')
                corr = float(tmp[7])
                for t3 in tmp3:
                    gene2 = t3.split('_')
                    # peak 2 belongs to gene 1 or are close in distance -> peak 2 is added to gene 1
                    if ((gene1 == gene2[0]) or (float(tmp[11]) < peak_dist)):
                        gene = gene1
                        if gene in corr_dict:
                            if peak2 not in corr_dict[gene] and gene2[1] != "promoter" and (gene2[0] not in gene_body_dict or peak2 not in gene_body_dict[gene2[0]]):
                                corr_dict[gene][0].append(peak2)
                                corr_dict[gene][1].append(corr)
                        else:
                            if gene2[1] != "promoter" and (gene2[0] not in gene_body_dict or peak2 not in gene_body_dict[gene2[0]]):
                                corr_dict[gene] = [[peak2], [corr]]

    gene_dict = promoter_dict
    enhancer_dict = {}
    promoter_genes = list(promoter_dict.keys())
    if verbose:
        print(f'Found {len(promoter_genes)} genes with promoter peaks')
    for gene in promoter_genes:
        if gene_body: # add gene-body peaks
            if gene in gene_body_dict:
                for peak in gene_body_dict[gene]:
                    if peak not in gene_dict[gene]:
                        gene_dict[gene].append(peak)
        enhancer_dict[gene] = []
        if gene in corr_dict: # add enhancer peaks
            for j, peak in enumerate(corr_dict[gene][0]):
                corr = corr_dict[gene][1][j]
                if corr > min_corr:
                    if peak not in gene_dict[gene]:
                        gene_dict[gene].append(peak)
                        enhancer_dict[gene].append(peak)

    # aggregate to genes
    adata_atac_X_copy = adata_atac.X.A
    gene_mat = np.zeros((adata_atac.shape[0], len(promoter_genes)))
    var_names = adata_atac.var_names.to_numpy()
    for i, gene in tqdm(enumerate(promoter_genes), total=len(promoter_genes)):
        peaks = gene_dict[gene]
        for peak in peaks:
            if peak in var_names:
                peak_index = np.where(var_names == peak)[0][0]
                gene_mat[:,i] += adata_atac_X_copy[:,peak_index]
    gene_mat[gene_mat < 0] = 0
    gene_mat = AnnData(X=sp.csr_matrix(gene_mat), dtype=np.float32)
    gene_mat.obs_names = pd.Index(list(adata_atac.obs_names))
    gene_mat.var_names = pd.Index(promoter_genes)
    gene_mat = gene_mat[:,gene_mat.X.sum(0) > 0]
    if return_dict:
        return gene_mat, promoter_dict, enhancer_dict
    else:
        return gene_mat
    
def tfidf_norm(adata_atac, scale_factor=1e4, copy=False):
    """TF-IDF normalization.

    This function normalizes counts in an AnnData object with TF-IDF.

    Parameters
    ----------
    adata_atac: :class:`~anndata.AnnData`
        ATAC anndata object.
    scale_factor: `float` (default: 1e4)
        Value to be multiplied after normalization.
    copy: `bool` (default: `False`)
        Whether to return a copy or modify `.X` directly.

    Returns
    -------
    If `copy==True`, a new ATAC anndata object which stores normalized counts
    in `.X`.
    """
    npeaks = adata_atac.X.sum(1)
    npeaks_inv = sp.csr_matrix(1.0/npeaks)
    tf = adata_atac.X.multiply(npeaks_inv)
    idf = sp.diags(np.ravel(adata_atac.X.shape[0] / adata_atac.X.sum(0))).log1p()
    if copy:
        adata_atac_copy = adata_atac.copy()
        adata_atac_copy.X = tf.dot(idf) * scale_factor
        return adata_atac_copy
    else:
        adata_atac.X = tf.dot(idf) * scale_factor

def lsi(
        adata: AnnData, n_components: int = 20, layer: Optional[str] = None,
        use_highly_variable: Optional[bool] = None, **kwargs
) -> None:
    r"""
    LSI analysis (following the Seurat v3 approach)
    Parameters
    ----------
    adata
        Input dataset
    n_components
        Number of dimensions to use
    layer
        Layer key to extract count data froma
    use_highly_variable
        Whether to use highly variable features only, stored in
        ``adata.var['highly_variable']``. By default uses them if they
        have been determined beforehand.
    **kwargs
        Additional keyword arguments are passed to
        :func:`sklearn.utils.extmath.randomized_svd`
    """
    if "random_state" not in kwargs:
        kwargs["random_state"] = 0  # Keep deterministic as the default behavior
    if use_highly_variable is None:
        use_highly_variable = "highly_variable" in adata.var
    adata_use = adata[:, adata.var["highly_variable"]] if use_highly_variable else adata
    X = adata_use.X if layer is None else adata_use.layers[layer]
    X_norm = preprocessing.normalize(X, norm="l1")
    X_norm = np.log1p(X_norm * 1e4)
    U, Sigma, VT = sklearn.utils.extmath.randomized_svd(X_norm, n_components, **kwargs)
    X_lsi = U - U.mean(axis=1, keepdims=True)
    X_lsi /= X_lsi.std(axis=1, ddof=1, keepdims=True)
    adata.obsm["X_lsi"] = X_lsi.astype(np.float32)

    adata.uns['lsi'] = {}
    lsi_dict = adata.uns['lsi']
    lsi_dict['U'] = U
    lsi_dict['Sigma'] = Sigma
    lsi_dict['VT'] = VT
    lsi_dict['params'] = {'n_components': n_components}


# WNN related
# pyWNN is a package developed by Dylan Kotliar (GitHub username: dylkot), published under the MIT license.
# The original release, including tutorials, can be found here: https://github.com/dylkot/pyWNN
def get_nearestneighbor(knn, neighbor=1):
    '''For each row of knn, returns the column with the lowest value
    I.e. the nearest neighbor'''
    indices = knn.indices
    indptr = knn.indptr
    data = knn.data
    nn_idx = []
    for i in range(knn.shape[0]):
        cols = indices[indptr[i]:indptr[i+1]]
        rowvals = data[indptr[i]:indptr[i+1]]
        idx = np.argsort(rowvals)
        nn_idx.append(cols[idx[neighbor-1]])
    return(np.array(nn_idx))


def compute_bw(knn_adj, embedding, n_neighbors=20):
    intersect = knn_adj.dot(knn_adj.T)
    indices = intersect.indices
    indptr = intersect.indptr
    data = intersect.data
    data = data / ((n_neighbors*2) - data)
    bandwidth = []
    for i in range(intersect.shape[0]):
        cols = indices[indptr[i]:indptr[i+1]]
        rowvals = data[indptr[i]:indptr[i+1]]
        idx = np.argsort(rowvals)
        valssort = rowvals[idx]
        numinset = len(cols)
        if numinset<n_neighbors:
            sys.exit('Fewer than 20 cells with Jacard sim > 0')
        else:
            curval = valssort[n_neighbors]
            for num in range(n_neighbors, numinset):
                if valssort[num]!=curval:
                    break
                else:
                    num+=1
            minjacinset = cols[idx][:num]
            if num <n_neighbors:
                print('shouldnt end up here')
                sys.exit(-1)
            else:
                euc_dist = ((embedding[minjacinset,:]-embedding[i,:])**2).sum(axis=1)**.5
                euc_dist_sorted = np.sort(euc_dist)[::-1]
                bandwidth.append( np.mean(euc_dist_sorted[:n_neighbors]) )
    return(np.array(bandwidth))


def compute_affinity(dist_to_predict, dist_to_nn, bw):
    affinity = dist_to_predict-dist_to_nn
    affinity[affinity<0]=0
    affinity = affinity * -1
    affinity = np.exp(affinity / (bw-dist_to_nn))
    return(affinity)


def dist_from_adj(adjacency, embed1, embed2, nndist1, nndist2):
    dist1 = sp.lil_matrix(adjacency.shape)
    dist2 = sp.lil_matrix(adjacency.shape)

    count = 0
    indices = adjacency.indices
    indptr = adjacency.indptr
    ncells = adjacency.shape[0]

    tic = time.perf_counter()
    for i in range(ncells):
        for j in range(indptr[i], indptr[i+1]):
            col = indices[j]
            a = (((embed1[i,:] - embed1[col,:])**2).sum()**.5) - nndist1[i]
            if a == 0: dist1[i,col] = np.nan
            else: dist1[i,col] = a
            b = (((embed2[i,:] - embed2[col,:])**2).sum()**.5) - nndist2[i]
            if b == 0: dist2[i,col] = np.nan
            else: dist2[i,col] = b

        if (i % 2000) == 0:
            toc = time.perf_counter()
            print('%d out of %d %.2f seconds elapsed' % (i, ncells, toc-tic))

    return(sp.csr_matrix(dist1), sp.csr_matrix(dist2))


def select_topK(dist,  n_neighbors=20):
    indices = dist.indices
    indptr = dist.indptr
    data = dist.data
    nrows = dist.shape[0]

    final_data = []
    final_col_ind = []

    tic = time.perf_counter()
    for i in range(nrows):
        cols = indices[indptr[i]:indptr[i+1]]
        rowvals = data[indptr[i]:indptr[i+1]]
        idx = np.argsort(rowvals)
        final_data.append(rowvals[idx[(-1*n_neighbors):]])
        final_col_ind.append(cols[idx[(-1*n_neighbors):]])

    final_data = np.concatenate(final_data)
    final_col_ind = np.concatenate(final_col_ind)
    final_row_ind = np.tile(np.arange(nrows), (n_neighbors, 1)).reshape(-1, order='F')

    result = sp.csr_matrix((final_data, (final_row_ind, final_col_ind)), shape=(nrows, dist.shape[1]))

    return(result)


class pyWNN():

    def __init__(self, adata, reps=['X_pca', 'X_apca'], n_neighbors=30, npcs=[20, 20], seed=0, distances=None):
        """
        Class for running weighted nearest neighbors analysis as described in Hao
        et al 2021.
        """

        self.seed = seed
        np.random.seed(seed)

        if len(reps)>2:
            sys.exit('WNN currently only implemented for 2 modalities')

        self.adata = adata.copy()
        self.reps = [r+'_norm' for r in reps]
        self.npcs = npcs
        for (i,r) in enumerate(reps):
            self.adata.obsm[self.reps[i]] = preprocessing.normalize(adata.obsm[r][:,0:npcs[i]])

        self.n_neighbors = n_neighbors
        if distances is None:
            sc.pp.neighbors(self.adata, n_neighbors=n_neighbors, n_pcs=npcs[0], use_rep=self.reps[0], metric='euclidean', key_added='1')
            sc.pp.neighbors(self.adata, n_neighbors=n_neighbors, n_pcs=npcs[1], use_rep=self.reps[1], metric='euclidean', key_added='2')
            sc.pp.neighbors(self.adata, n_neighbors=200, n_pcs=npcs[0], use_rep=self.reps[0], metric='euclidean', key_added='1_200')
            sc.pp.neighbors(self.adata, n_neighbors=200, n_pcs=npcs[1], use_rep=self.reps[1], metric='euclidean', key_added='2_200')
            self.distances = ['1_distances', '2_distances', '1_200_distances', '2_200_distances']
        else:
            self.distances = distances

        for d in self.distances:
            if type(self.adata.obsp[d]) is not sp.csr_matrix:
                self.adata.obsp[d] = sp.csr_matrix(self.adata.obsp[d])

        self.NNdist = []
        self.NNidx = []
        self.NNadjacency = []
        self.BWs = []

        for (i,r) in enumerate(self.reps):
            nn = get_nearestneighbor(self.adata.obsp[self.distances[i]])
            dist_to_nn = ((self.adata.obsm[r]-self.adata.obsm[r][nn, :])**2).sum(axis=1)**.5
            nn_adj = (self.adata.obsp[self.distances[i]]>0).astype(int)
            nn_adj_wdiag = nn_adj.copy()
            nn_adj_wdiag.setdiag(1)
            bw = compute_bw(nn_adj_wdiag, self.adata.obsm[r], n_neighbors=self.n_neighbors)
            self.NNidx.append(nn)
            self.NNdist.append(dist_to_nn)
            self.NNadjacency.append(nn_adj)
            self.BWs.append(bw)

        self.weights = []
        self.WNN = None

    def compute_weights(self):
        cmap = {0:1, 1:0}
        affinity_ratios = []
        self.within = []
        self.cross = []
        for (i,r) in enumerate(self.reps):
            within_predict = self.NNadjacency[i].dot(self.adata.obsm[r]) / (self.n_neighbors-1)
            cross_predict = self.NNadjacency[cmap[i]].dot(self.adata.obsm[r]) / (self.n_neighbors-1)

            within_predict_dist = ((self.adata.obsm[r] - within_predict)**2).sum(axis=1)**.5
            cross_predict_dist = ((self.adata.obsm[r] - cross_predict)**2).sum(axis=1)**.5
            within_affinity = compute_affinity(within_predict_dist, self.NNdist[i], self.BWs[i])
            cross_affinity = compute_affinity(cross_predict_dist, self.NNdist[i], self.BWs[i])
            affinity_ratios.append(within_affinity / (cross_affinity + 0.0001))
            self.within.append(within_predict_dist)
            self.cross.append(cross_predict_dist)

        self.weights.append( 1 / (1+ np.exp(affinity_ratios[1]-affinity_ratios[0])) )
        self.weights.append( 1 - self.weights[0] )


    def compute_wnn(self, adata):
        self.compute_weights()
        union_adj_mat = ((self.adata.obsp[self.distances[2]]+self.adata.obsp[self.distances[3]]) > 0).astype(int)

        full_dists = dist_from_adj(union_adj_mat, self.adata.obsm[self.reps[0]], self.adata.obsm[self.reps[1]],
                                   self.NNdist[0], self.NNdist[1])
        weighted_dist = sp.csr_matrix(union_adj_mat.shape)
        for (i,dist) in enumerate(full_dists):
            dist = sp.diags(-1 / (self.BWs[i] - self.NNdist[i]), format='csr').dot(dist)
            dist.data = np.exp(dist.data)
            ind = np.isnan(dist.data)
            dist.data[ind] = 1
            dist = sp.diags(self.weights[i]).dot(dist)
            weighted_dist += dist

        self.WNN = select_topK(weighted_dist,  n_neighbors=self.n_neighbors)
        WNNdist = self.WNN.copy()
        x = (1-WNNdist.data) / 2
        x[x<0]=0
        x[x>1]=1
        WNNdist.data = np.sqrt(x)
        self.WNNdist = WNNdist

        adata.obsp['WNN'] = self.WNN
        adata.obsp['WNN_distance'] = self.WNNdist
        adata.obsm[self.reps[0]] = self.adata.obsm[self.reps[0]]
        adata.obsm[self.reps[1]] = self.adata.obsm[self.reps[1]]
        adata.uns['WNN'] = {'connectivities_key': 'WNN',
                                     'distances_key': 'WNN_distance',
                                     'params': {'n_neighbors': self.n_neighbors,
                                      'method': 'WNN',
                                      'random_state': self.seed,
                                      'metric': 'euclidean',
                                      'use_rep': self.reps[0],
                                      'n_pcs': self.npcs[0]}}
        return(adata)
    

def gen_wnn(
    adata,
    g_basis='X_pca',
    c_basis='X_lsi', 
    k=30, 
    dims=None, 
    copy=False, 
    random_state=0):
    """Computes inputs for KNN smoothing.

    This function calculates the nn_idx and nn_dist matrices needed
    to run knn_smooth_chrom().
    Adapted from MultiVelo.

    Parameters
    ----------
    rna: :class:`~anndata.AnnData`
        RNA anndata object.
    adata_atac: :class:`~anndata.AnnData`
        ATAC anndata object.
    dims: `List[int]`
        Dimensions of data for RNA (index=0) and ATAC (index=1)
    k: `int` (default: `None`)
        Top N neighbors to extract for each cell in the connectivities matrix.

    Returns
    -------
    nn_idx: `np.darray` (default: `None`)
        KNN index matrix of size (cells, k).
    nn_dist: `np.darray` (default: `None`)
        KNN distance matrix of size (cells, k).
    """
    if copy:
        adata = adata.copy()
    if dims is None:
        dims = [adata.obsm[g_basis].shape[1], adata.obsm[c_basis].shape[1]]

    # run WNN
    WNNobj = pyWNN(adata,
                      reps=[g_basis, c_basis],
                      npcs=dims,
                      n_neighbors=k,
                      seed=random_state)
    adata = WNNobj.compute_wnn(adata)

    # get the matrix storing the distances between each cell and its neighbors
    cx = scipy.sparse.coo_matrix(adata.obsp["WNN_distance"])

    # the number of cells
    cells = adata.obsp['WNN_distance'].shape[0]

    # define the shape of our final results
    # and make the arrays that will hold the results
    new_shape = (cells, k)
    nn_dist = np.zeros(shape=new_shape)
    nn_idx = np.zeros(shape=new_shape)

    # new_col defines what column we store data in
    # our result arrays
    new_col = 0

    # loop through the distance matrices
    for i, j, v in zip(cx.row, cx.col, cx.data):

        # store the distances between neighbor cells
        nn_dist[i][new_col % k] = v

        # for each cell's row, store the row numbers of its neighbor cells
        # (1-indexing instead of 0- is a holdover from R multimodalneighbors())
        nn_idx[i][new_col % k] = int(j)

        new_col += 1

    adata.uns['WNN']['indices'] = nn_idx.astype(int)

    from umap.umap_ import fuzzy_simplicial_set
    X_tmp = sp.coo_matrix(([], ([], [])), shape=(nn_idx.shape[0], 1))
    conn, sigmas, rhos, dists = fuzzy_simplicial_set(X_tmp, nn_idx.shape[1],
                                                    None, None,
                                                    knn_indices=nn_idx,
                                                    knn_dists=nn_dist,
                                                    return_dists=True)
    conn = conn.tocsr().copy()
    conn.setdiag(1)
    adata.obsp['WNN_connectivities'] = conn

    return adata


from typing import Tuple, Optional, Union, List
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

import logging
from tqdm import tqdm
import numpy as np
import pandas as pd
import scipy.sparse as sp
from pandas.api.types import is_numeric_dtype, is_categorical_dtype
from pygam import LinearGAM, s

import matplotlib
import matplotlib.pyplot as plt
from matplotlib.colors import Colormap
import matplotlib.patheffects as PathEffects
import seaborn as sns

from .utils import flatten
from .gam import fit_velo_peak
from .metrics import cross_boundary_correctness
from .kernel_density_smooth import kde2d, kde2d_to_mean_and_sigma


# Cell cycle genes are adapted from `dynamo`.
G1S_genes_human = [
        "ARGLU1",
        "BRD7",
        "CDC6",
        "CLSPN",
        "ESD",
        "GINS2",
        "GMNN",
        "LUC7L3",
        "MCM5",
        "MCM6",
        "NASP",
        "PCNA",
        "PNN",
        "SLBP",
        "SRSF7",
        "SSR3",
        "ZRANB2"]
S_genes_human = [
        "ASF1B",
        "CALM2",
        "CDC45",
        "CDCA5",
        "CENPM",
        "DHFR",
        "EZH2",
        "FEN1",
        "HIST1H2AC",
        "HIST1H4C",
        "NEAT1",
        "PKMYT1",
        "PRIM1",
        "RFC2",
        "RPA2",
        "RRM2",
        "RSRC2",
        "SRSF5",
        "SVIP",
        "TOP2A",
        "TYMS",
        "UBE2T",
        "ZWINT"]
G2M_genes_human = [
        "AURKB",
        "BUB3",
        "CCNA2",
        "CCNF",
        "CDCA2",
        "CDCA3",
        "CDCA8",
        "CDK1",
        "CKAP2",
        "DCAF7",
        "HMGB2",
        "HN1",
        "KIF5B",
        "KIF20B",
        "KIF22",
        "KIF23",
        "KIFC1",
        "KPNA2",
        "LBR",
        "MAD2L1",
        "MALAT1",
        "MND1",
        "NDC80",
        "NUCKS1",
        "NUSAP1",
        "PIF1",
        "PSMD11",
        "PSRC1",
        "SMC4",
        "TIMP1",
        "TMEM99",
        "TOP2A",
        "TUBB",
        "TUBB4B",
        "VPS25"]
M_genes_human = [
        "ANP32B",
        "ANP32E",
        "ARL6IP1",
        "AURKA",
        "BIRC5",
        "BUB1",
        "CCNA2",
        "CCNB2",
        "CDC20",
        "CDC27",
        "CDC42EP1",
        "CDCA3",
        "CENPA",
        "CENPE",
        "CENPF",
        "CKAP2",
        "CKAP5",
        "CKS1B",
        "CKS2",
        "DEPDC1",
        "DLGAP5",
        "DNAJA1",
        "DNAJB1",
        "GRK6",
        "GTSE1",
        "HMG20B",
        "HMGB3",
        "HMMR",
        "HN1",
        "HSPA8",
        "KIF2C",
        "KIF5B",
        "KIF20B",
        "LBR",
        "MKI67",
        "MZT1",
        "NUF2",
        "NUSAP1",
        "PBK",
        "PLK1",
        "PRR11",
        "PSMG3",
        "PWP1",
        "RAD51C",
        "RBM8A",
        "RNF126",
        "RNPS1",
        "RRP1",
        "SFPQ",
        "SGOL2",
        "SMARCB1",
        "SRSF3",
        "TACC3",
        "THRAP3",
        "TPX2",
        "TUBB4B",
        "UBE2D3",
        "USP16",
        "WIBG",
        "YWHAH",
        "ZNF207"]
MG1_genes_human = [
        "AMD1",
        "ANP32E",
        "CBX3",
        "CDC42",
        "CNIH4",
        "CWC15",
        "DKC1",
        "DNAJB6",
        "DYNLL1",
        "EIF4E",
        "FXR1",
        "GRPEL1",
        "GSPT1",
        "HMG20B",
        "HSPA8",
        "ILF2",
        "KIF5B",
        "KPNB1",
        "LARP1",
        "LYAR",
        "MORF4L2",
        "MRPL19",
        "MRPS2",
        "MRPS18B",
        "NUCKS1",
        "PRC1",
        "PTMS",
        "PTTG1",
        "RAN",
        "RHEB",
        "RPL13A",
        "SRSF3",
        "SYNCRIP",
        "TAF9",
        "TMEM138",
        "TOP1",
        "TROAP",
        "UBE2D3",
        "ZNF593"]

def polar_plot(adata, phase_key='phase', layer='velocity_gv', show_names=False, show_markers=True, show=False, **kwargs):
    """
    Polar plot adapted from VeloCycle(https://github.com/lamanno-epfl/velocycle/blob/main/velocycle/plots.py) with some modifications.
    """
    gene_names = np.array([a for a in adata.var_names])
    markers = G1S_genes_human + S_genes_human + G2M_genes_human + M_genes_human + MG1_genes_human
    phases_list = [G1S_genes_human, S_genes_human, G2M_genes_human, M_genes_human, MG1_genes_human, [i for i in gene_names if i.upper() not in markers]]
    gs = []
    gradient = []
    for i in range(len(phases_list)):
        for j in range(len(phases_list[i])):
            if phases_list[i][j] not in gene_names:
                continue
            else:
                gs.append(phases_list[i][j])
                gradient.append(i)

    color_gradient_map = pd.DataFrame({'Gene': gs,  'Color': gradient}).set_index('Gene').to_dict()['Color']
    colored_gradient = pd.Series(gs).map(color_gradient_map)
    gs = np.array(gs)
    
    for i,j in zip(gs, colored_gradient):
        if np.isnan(j):
            print(i)
         
    # rescale phase for polar plot
    phase = adata.obs[phase_key].values
    phase -= phase.min()
    phase = phase / phase.max() * 2 * np.pi

    gam_kwargs = {
        'n_splines': 7,
        'spline_order': 3,
    }
    gam_kwargs.update(kwargs)

    df_peak = fit_velo_peak(
        adata, 
        genes=gs, 
        tkey=phase_key, 
        layer=layer, 
        log_norm=True, 
        max_iter=1000, 
        **gam_kwargs
    )
    
    # Extract the peak phase and magnitude values.
    angle = df_peak['phase'].values
    magnitude = df_peak['magnitude'].values
    
    fig = plt.figure(figsize = (6, 6))
    ax = fig.add_subplot(projection='polar')
    
    # First: only plot dots with a color assignment
    angle_subset = angle[~np.isnan(colored_gradient.values)]
    r_subset = magnitude[~np.isnan(colored_gradient.values)]
    color_subset = colored_gradient.values[~np.isnan(colored_gradient.values)]
    
    # Remove genes with negative velo
    angle_subset = angle_subset[r_subset>=0]
    color_subset = color_subset[r_subset>=0]
    gene_names_subset = gs[r_subset>=0]
    r_subset = r_subset[r_subset>=0]
    
    # Plot all genes in phases list
    cell_cycle_cmap = {0:'firebrick', 1:'orange', 2:'yellowgreen', 3:'teal', 4:'royalblue', 5: 'black'} 
    ax.scatter(angle_subset, r_subset, c=[cell_cycle_cmap[i] for i in color_subset], s=50, alpha=0.2, edgecolor='none', rasterized=True)
    
    angle_subset_markers = angle_subset[color_subset!=5]
    r_subset_markers = r_subset[color_subset!=5]
    gene_names_subset_markers = gene_names_subset[color_subset!=5]
    color_subset_markers = color_subset[color_subset!=5]
    
    ax.scatter(angle_subset_markers, r_subset_markers, c=[cell_cycle_cmap[i] for i in color_subset_markers], s=80, alpha=1, edgecolor='none',rasterized=True)
    
    # Annotate genes
    if show_markers:
        for (i, txt), c in zip(enumerate(gene_names_subset_markers), colored_gradient.values):
            ix = np.where(np.array(gene_names_subset_markers)==txt)[0][0]
            ax.annotate(txt[0]+txt[1:].upper(), (angle_subset_markers[ix], r_subset_markers[ix]+0.02))

    if show_names:
        for (i, txt), c in zip(enumerate(gs), colored_gradient.values):
            ix = np.where(gs==txt)[0][0]
            ax.annotate(txt[0]+txt[1:].upper(), (angle[ix], magnitude[ix]+0.02))

    plt.xlim(0, 2*np.pi)
    plt.ylim(-1, )
    plt.yticks([-1, -0.5, 0, 0.5, 1])
    plt.xticks([0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi],["0", "π/2", "π", "3π/2", "2π"])
    plt.tight_layout()
    if show:
        plt.show();
    else: 
        return ax


def cbc_heatmap(
        adata,
        xkey:str='Ms',
        vkey:str='velocity',
        cluster_key:str='clusters', 
        basis:str='pca',
        neighbor_key:str='neighbors',
        vector:str='velocity', 
        corr_func='cosine',
        cmap = 'bwr',
        annot: bool = True,
        custom_order: list = None,
        ):
    """
    Visualize the CBC score of every possible transitions between clusters using heatmap. 
    """
    clusters = adata.obs[cluster_key].unique()
    rows, cols = clusters.to_list(), clusters.to_list()
    cluster_edges = []
    for i in clusters:
        for j in clusters:
            if i == j:
                continue
            else:
                cluster_edges.append((i, j))

    scores, _ = cross_boundary_correctness(
        adata, xkey=xkey, vkey=vkey, cluster_key=cluster_key, cluster_edges=cluster_edges,
        basis=basis, neighbor_key=neighbor_key, vector=vector, corr_func=corr_func, return_raw=False)

    df = pd.DataFrame(index=rows, columns=cols)
    for row in rows:
        for col in cols:
            if row == col:
                df.loc[row, col] = np.nan
            else:
                df.loc[row, col] = scores[(row, col)]

    df = df.apply(pd.to_numeric)
    if custom_order:
        df = df.reindex(index=custom_order, columns=custom_order)
        
    abs_max = np.nanmax(np.abs(df.values))

    sns.set_style("whitegrid")
    sns.heatmap(df, cmap=cmap, annot=annot, center=0, vmin=-abs_max, vmax=abs_max)
    plt.show()


def scatter_plot_mm(adata,
                 genes,
                 vukey,
                 vskey,
                 vckey,
                 by='cus',
                 color_by='celltype',
                 n_cols=5,
                 axis_on=True,
                 frame_on=True,
                 velocity_arrows=False,
                 downsample=1,
                 figsize=None,
                 pointsize=2,
                 cmap='coolwarm',
                 view_3d_elev=None,
                 view_3d_azim=None,
                 full_name=False
                 ):
    """Gene scatter plot.

    This function plots phase portraits of the specified plane.
    Modification of Multivelo

    Parameters
    ----------
    adata: :class:`~anndata.AnnData`
        Anndata result from dynamics recovery.
    genes: `str`,  list of `str`
        List of genes to plot.
    by: `str` (default: `cus`)
        Plot unspliced-spliced plane if `us`. Plot chromatin-unspliced plane
        if `cu`.
        Plot 3D phase portraits if `cus`.
    color_by: `str` (default: `state`)
        Color by the cell labels like leiden, louvain, celltype, etc.
        The color field must be present in `.uns`, which can be pre-computed 
        with `scanpy.pl.scatter`.
        When `by=='us'`, `color_by` can also be `c`, which displays the log
        accessibility on U-S phase portraits.
    n_cols: `int` (default: 5)
        Number of columns to plot on each row.
    axis_on: `bool` (default: `True`)
        Whether to show axis labels.
    frame_on: `bool` (default: `True`)
        Whether to show plot frames.
    title_more_info: `bool` (default: `False`)
        Whether to display model, direction, and likelihood information for
        the gene in title.
    velocity_arrows: `bool` (default: `False`)
        Whether to show velocity arrows of cells on the phase portraits.
    downsample: `int` (default: 1)
        How much to downsample the cells. The remaining number will be
        `1/downsample` of original.
    figsize: `tuple` (default: `None`)
        Total figure size.
    pointsize: `float` (default: 2)
        Point size for scatter plots.
    markersize: `float` (default: 5)
        Point size for switch time points.
    linewidth: `float` (default: 2)
        Line width for connected anchors.
    cmap: `str` (default: `coolwarm`)
        Color map for log accessibilities or other continuous color keys when
        plotting on U-S plane.
    view_3d_elev: `float` (default: `None`)
        Matplotlib 3D plot `elev` argument. `elev=90` is the same as U-S plane,
        and `elev=0` is the same as C-U plane.
    view_3d_azim: `float` (default: `None`)
        Matplotlib 3D plot `azim` argument. `azim=270` is the same as U-S
        plane, and `azim=0` is the same as C-U plane.
    full_name: `bool` (default: `False`)
        Show full names for chromatin, unspliced, and spliced rather than
        using abbreviated terms c, u, and s.
    """
    if by not in ['us', 'cu', 'cus']:
        raise ValueError("'by' argument must be one of ['us', 'cu', 'cus']")
    if by == 'us' and color_by == 'c':
        cell_annot = None
    elif color_by in adata.obs and is_numeric_dtype(adata.obs[color_by]):
        cell_annot = None
        colors = adata.obs[color_by].values
    elif color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]) \
            and color_by+'_colors' in adata.uns.keys():
        cell_annot = adata.obs[color_by].cat.categories
        if isinstance(adata.uns[f'{color_by}_colors'], dict):
            colors = list(adata.uns[f'{color_by}_colors'].values())
        else:
            colors = adata.uns[f'{color_by}_colors']
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata'
                         ' obs, and the colors of categories can be found in'
                         ' adata uns.')

    downsample = np.clip(int(downsample), 1, 10)
    genes = np.array(genes)
    missing_genes = genes[~np.isin(genes, adata.var_names)]
    if len(missing_genes) > 0:
        print(f'{missing_genes} not found')
    genes = genes[np.isin(genes, adata.var_names)]
    gn = len(genes)
    if gn == 0:
        return
    if gn < n_cols:
        n_cols = gn
    if by == 'cus':
        fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(3.2*n_cols, 2.7*(-(-gn // n_cols)))
                                if figsize is None else figsize,
                                subplot_kw={'projection': '3d'})
    else:
        fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(2.7*n_cols, 2.4*(-(-gn // n_cols)))
                                if figsize is None else figsize)
    fig.patch.set_facecolor('white')
    count = 0
    for gene in genes:
        u = adata[:, gene].layers['Mu'].copy() if 'Mu' in adata.layers \
            else adata[:, gene].layers['unspliced'].copy()
        s = adata[:, gene].layers['Ms'].copy() if 'Ms' in adata.layers \
            else adata[:, gene].layers['spliced'].copy()
        u = u.A if sp.issparse(u) else u
        s = s.A if sp.issparse(s) else s
        u, s = np.ravel(u), np.ravel(s)
        if 'ATAC' not in adata.layers.keys() and \
                'Mc' not in adata.layers.keys():
            raise ValueError('Cannot find ATAC data in adata layers.')
        elif 'ATAC' in adata.layers.keys():
            c = adata[:, gene].layers['ATAC'].copy()
            c = c.A if sp.issparse(c) else c
            c = np.ravel(c)
        elif 'Mc' in adata.layers.keys():
            c = adata[:, gene].layers['Mc'].copy()
            c = c.A if sp.issparse(c) else c
            c = np.ravel(c)

        if velocity_arrows:
            if vukey in adata.layers.keys():
                vu = adata[:, gene].layers[vukey].copy()
            else:
                vu = np.zeros(adata.n_obs)
            max_u = np.max([np.max(u), 1e-6])
            u /= max_u
            vu = np.ravel(vu)
            vu /= np.max([np.max(np.abs(vu)), 1e-6])
            if vskey in adata.layers.keys():
                vs = adata[:, gene].layers[vskey].copy()
            else:
                raise ValueError(f'Splicing velocity {vskey} can not be found in adata layers.')
            max_s = np.max([np.max(s), 1e-6])
            s /= max_s
            vs = np.ravel(vs)
            vs /= np.max([np.max(np.abs(vs)), 1e-6])
            if vckey in adata.layers.keys():
                vc = adata[:, gene].layers[vckey].copy()
                max_c = np.max([np.max(c), 1e-6])
                c /= max_c
                vc = np.ravel(vc)
                vc /= np.max([np.max(np.abs(vc)), 1e-6])

        row = count // n_cols
        col = count % n_cols
        ax = axs[row, col]
        if cell_annot is not None:
            for i in range(len(cell_annot)):
                filt = adata.obs[color_by] == cell_annot[i]
                filt = np.ravel(filt)
                if by == 'us':
                    if velocity_arrows:
                        ax.quiver(s[filt][::downsample], u[filt][::downsample],
                                  vs[filt][::downsample],
                                  vu[filt][::downsample], color=colors[i],
                                  alpha=0.5, scale_units='xy', scale=10,
                                  width=0.005, headwidth=4, headaxislength=5.5)
                    else:
                        ax.scatter(s[filt][::downsample],
                                   u[filt][::downsample], s=pointsize,
                                   c=colors[i], alpha=0.7)
                elif by == 'cu':
                    if velocity_arrows:
                        ax.quiver(u[filt][::downsample],
                                  c[filt][::downsample],
                                  vu[filt][::downsample],
                                  vc[filt][::downsample], color=colors[i],
                                  alpha=0.5, scale_units='xy', scale=10,
                                  width=0.005, headwidth=4, headaxislength=5.5)
                    else:
                        ax.scatter(u[filt][::downsample],
                                   c[filt][::downsample], s=pointsize,
                                   c=colors[i], alpha=0.7)
                else:
                    if velocity_arrows:
                        ax.quiver(s[filt][::downsample],
                                  u[filt][::downsample], c[filt][::downsample],
                                  vs[filt][::downsample],
                                  vu[filt][::downsample],
                                  vc[filt][::downsample],
                                  color=colors[i], alpha=0.4, length=0.1,
                                  arrow_length_ratio=0.5, normalize=True)
                    else:
                        ax.scatter(s[filt][::downsample],
                                   u[filt][::downsample],
                                   c[filt][::downsample], s=pointsize,
                                   c=colors[i], alpha=0.7)
        elif color_by == 'c':
            outlier = 99.8
            non_zero = (u > 0) & (s > 0) & (c > 0)
            non_outlier = u < np.percentile(u, outlier)
            non_outlier &= s < np.percentile(s, outlier)
            non_outlier &= c < np.percentile(c, outlier)
            c -= np.min(c)
            c /= np.max(c)
            if velocity_arrows:
                ax.quiver(s[non_zero & non_outlier][::downsample],
                          u[non_zero & non_outlier][::downsample],
                          vs[non_zero & non_outlier][::downsample],
                          vu[non_zero & non_outlier][::downsample],
                          np.log1p(c[non_zero & non_outlier][::downsample]),
                          alpha=0.5,
                          scale_units='xy', scale=10, width=0.005,
                          headwidth=4, headaxislength=5.5, cmap=cmap)
            else:
                ax.scatter(s[non_zero & non_outlier][::downsample],
                           u[non_zero & non_outlier][::downsample],
                           s=pointsize,
                           c=np.log1p(c[non_zero & non_outlier][::downsample]),
                           alpha=0.8, cmap=cmap)
        else:
            if by == 'us':
                if velocity_arrows:
                    ax.quiver(s[::downsample], u[::downsample],
                              vs[::downsample], vu[::downsample],
                              colors[::downsample], alpha=0.5,
                              scale_units='xy', scale=10, width=0.005,
                              headwidth=4, headaxislength=5.5, cmap=cmap)
                else:
                    ax.scatter(s[::downsample], u[::downsample], s=pointsize,
                               c=colors[::downsample], alpha=0.7, cmap=cmap)
            elif by == 'cu':
                if velocity_arrows:
                    ax.quiver(u[::downsample], c[::downsample],
                              vu[::downsample], vc[::downsample],
                              colors[::downsample], alpha=0.5,
                              scale_units='xy', scale=10, width=0.005,
                              headwidth=4, headaxislength=5.5, cmap=cmap)
                else:
                    ax.scatter(u[::downsample], c[::downsample], s=pointsize,
                               c=colors[::downsample], alpha=0.7, cmap=cmap)
            else:
                if velocity_arrows:
                    ax.quiver(s[::downsample], u[::downsample],
                              c[::downsample], vs[::downsample],
                              vu[::downsample], vc[::downsample],
                              colors[::downsample], alpha=0.4, length=0.1,
                              arrow_length_ratio=0.5, normalize=True,
                              cmap=cmap)
                else:
                    ax.scatter(s[::downsample], u[::downsample],
                               c[::downsample], s=pointsize,
                               c=colors[::downsample], alpha=0.7, cmap=cmap)

        if by == 'cus' and \
                (view_3d_elev is not None or view_3d_azim is not None):
            # US: elev=90, azim=270. CU: elev=0, azim=0.
            ax.view_init(elev=view_3d_elev, azim=view_3d_azim)
        title = gene
        ax.set_title(f'{title}', fontsize=11)
        if by == 'us':
            ax.set_xlabel('spliced' if full_name else 's')
            ax.set_ylabel('unspliced' if full_name else 'u')
        elif by == 'cu':
            ax.set_xlabel('unspliced' if full_name else 'u')
            ax.set_ylabel('chromatin' if full_name else 'c')
        elif by == 'cus':
            ax.set_xlabel('spliced' if full_name else 's')
            ax.set_ylabel('unspliced' if full_name else 'u')
            ax.set_zlabel('chromatin' if full_name else 'c')
        if by in ['us', 'cu']:
            if not axis_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)
            if not frame_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.set_frame_on(False)
        elif by == 'cus':
            if not axis_on:
                ax.set_xlabel('')
                ax.set_ylabel('')
                ax.set_zlabel('')
                ax.xaxis.set_ticklabels([])
                ax.yaxis.set_ticklabels([])
                ax.zaxis.set_ticklabels([])
            if not frame_on:
                ax.xaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.yaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.zaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.xaxis._axinfo['tick']['inward_factor'] = 0
                ax.xaxis._axinfo['tick']['outward_factor'] = 0
                ax.yaxis._axinfo['tick']['inward_factor'] = 0
                ax.yaxis._axinfo['tick']['outward_factor'] = 0
                ax.zaxis._axinfo['tick']['inward_factor'] = 0
                ax.zaxis._axinfo['tick']['outward_factor'] = 0
        count += 1
    for i in range(col+1, n_cols):
        fig.delaxes(axs[row, i])
    fig.tight_layout()


def scatter_twogenes_3d(adata,
                 x,
                 y,
                 z,
                 xkey='M_sc',
                 vkey='velocity_sc',
                 color_by='celltype',
                 n_cols=5,
                 axis_on=True,
                 frame_on=True,
                 velocity_arrows=False,
                 downsample=1,
                 figsize=None,
                 pointsize=2,
                 alpha=0.4,
                 cmap='coolwarm',
                 show_legend=False, 
                 view_3d_elev=None,
                 view_3d_azim=None
                 ):
    """Gene scatter plot.
    """
    if isinstance(x, str):
        x = [x]
    if isinstance(y, str):
        y = [y]
    if isinstance(z, str):
        z = [z]
    if xkey not in adata.layers.keys() or vkey not in adata.layers.keys():
        raise ValueError('Provided xkey or vkey not found in adata.layers!')
    if color_by == 'vz':
        cell_annot = None
    elif color_by in adata.obs and is_numeric_dtype(adata.obs[color_by]):
        cell_annot = None
        colors = adata.obs[color_by].values
        colors -= np.min(colors)
        colors /= np.max(colors)
    elif color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]) \
            and color_by+'_colors' in adata.uns.keys():
        cell_annot = adata.obs[color_by].cat.categories
        if isinstance(adata.uns[f'{color_by}_colors'], dict):
            colors = list(adata.uns[f'{color_by}_colors'].values())
        else:
            colors = adata.uns[f'{color_by}_colors']
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata'
                         ' obs, and the colors of categories can be found in'
                         ' adata uns.')

    downsample = np.clip(int(downsample), 1, 10)
    genes = np.array(list(set(x + y + z)))
    missing_genes = genes[~np.isin(genes, adata.var_names)]
    if len(missing_genes) > 0:
        print(f'{missing_genes} not found')
    genes = genes[np.isin(genes, adata.var_names)]
    gn = len(x)
    if gn < n_cols:
        n_cols = gn
    if color_by == 'vz':
        fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(3*n_cols, 2.4*(-(-gn // n_cols)))
                                if figsize is None else figsize)
    else:
        fig, axs = plt.subplots(-(-gn // n_cols), n_cols, squeeze=False,
                                figsize=(3.8*n_cols, 2.7*(-(-gn // n_cols)))
                                if figsize is None else figsize,
                                subplot_kw={'projection': '3d'})
    point_size = 500.0 / np.sqrt(adata.shape[0]) if pointsize is None else 500.0 / np.sqrt(adata.shape[0]) * pointsize
    point_size = 4 * point_size

    fig.patch.set_facecolor('white')
    count = 0
    for a, b, c in zip(x, y, z):
        xa = adata[:, a].layers[xkey].copy()
        xb = adata[:, b].layers[xkey].copy()
        xc = adata[:, c].layers[xkey].copy()
        xa, xb, xc = np.ravel(xa), np.ravel(xb), np.ravel(xc)

        va = adata[:, a].layers[vkey].copy()
        vb = adata[:, b].layers[vkey].copy()
        vc = adata[:, c].layers[vkey].copy()
        max_a = np.max([np.max(xa), 1e-6])
        xa /= max_a
        va = np.ravel(va)
        va /= np.max([np.max(np.abs(va)), 1e-6])
        max_b = np.max([np.max(xb), 1e-6])
        xb /= max_b
        vb = np.ravel(vb)
        vb /= np.max([np.max(np.abs(vb)), 1e-6])
        max_c = np.max([np.max(xc), 1e-6])
        xc /= max_c
        vc = np.ravel(vc)
        vc /= np.max([np.max(np.abs(vc)), 1e-6])
        if color_by == 'vz':
            colors = vc

        cur_df = pd.DataFrame({"x": xa, "y": xb, "z": xc, 
                               "dx": va, "dy": vb, "dz": vc})

        row = count // n_cols
        col = count % n_cols
        ax = axs[row, col]
        if cell_annot is not None:
            for i in range(len(cell_annot)):
                filt = adata.obs[color_by] == cell_annot[i]
                filt = np.ravel(filt)
                if velocity_arrows:
                    ax.quiver(xa[filt][::downsample],
                              xb[filt][::downsample], 
                              xc[filt][::downsample],
                              va[filt][::downsample],
                              vb[filt][::downsample],
                              vc[filt][::downsample],
                              color=colors[i], alpha=alpha, length=0.1,
                              arrow_length_ratio=0.3, normalize=True)
                else:
                    ax.scatter(xa[filt][::downsample],
                               xb[filt][::downsample],
                               xc[filt][::downsample], s=pointsize,
                               c=colors[i], alpha=alpha)

        elif color_by == 'vz':
            vc_max = np.percentile(np.abs(vc).max(), 95)
            if velocity_arrows:
                ax.quiver(xa[::downsample],
                          xb[::downsample],
                          va[::downsample],
                          vb[::downsample],
                          color=colors[::downsample],
                          alpha=alpha,
                          scale_units='xy', scale=10, width=0.005,
                          headwidth=4, headaxislength=5.5, cmap=cmap)
            else:
                ax.scatter(xa[::downsample],
                           xb[::downsample],
                           s=pointsize,
                           c=colors,
                           alpha=alpha, cmap=cmap,
                           vmin=-vc_max, vmax=vc_max)
        else:
            if velocity_arrows:
                ax.quiver(xa[::downsample],
                            xb[::downsample], 
                            xc[::downsample],
                            va[::downsample],
                            vb[::downsample],
                            vc[::downsample],
                            color=colors[::downsample], alpha=alpha, length=0.1,
                            arrow_length_ratio=0.3, normalize=True,
                            cmap=cmap)
            else:
                ax.scatter(xa[::downsample],
                            xb[::downsample],
                            xc[::downsample], s=pointsize,
                            c=colors[::downsample], alpha=alpha, cmap=cmap)

        if show_legend:
            # plt.subplots_adjust(right=0.95)
            if color_by == 'vz' or (color_by in adata.obs and is_numeric_dtype(adata.obs[color_by])):
                # Create a scalar mappable for numeric data
                sm = plt.cm.ScalarMappable(cmap=cmap, norm=plt.Normalize(vmin=np.min(colors), vmax=np.max(colors)))
                sm.set_array([])
                cbar = fig.colorbar(sm, ax=ax, orientation='vertical', pad=0.1, aspect=20)
            elif color_by in adata.obs and is_categorical_dtype(adata.obs[color_by]):
                # Create a color bar with discrete colors for categorical data
                from matplotlib.colors import ListedColormap
                cmap = ListedColormap(colors)
                bounds = np.arange(len(cell_annot)+1)
                norm = plt.Normalize(vmin=bounds.min(), vmax=bounds.max())
                sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
                sm.set_array(bounds)
                cbar = fig.colorbar(sm, ticks=np.arange(len(cell_annot)), ax=ax, orientation='vertical', pad=0.1, aspect=20)
                cbar.set_ticklabels(cell_annot)

        if (view_3d_elev is not None or view_3d_azim is not None):
            # US: elev=90, azim=270. CU: elev=0, azim=0.
            ax.view_init(elev=view_3d_elev, azim=view_3d_azim)
        title = f'{a}_{b}_{c}'
        ax.set_title(title, fontsize=11)
        if color_by == 'vz':
            if not axis_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.get_xaxis().set_visible(False)
                ax.get_yaxis().set_visible(False)
            else:
                ax.set_xlabel(a)
                ax.set_ylabel(b)
                ax.set_xticklabels(ax.get_yticklabels(), rotation=90, ha='right')
            if not frame_on:
                ax.xaxis.set_ticks_position('none')
                ax.yaxis.set_ticks_position('none')
                ax.set_frame_on(False)
        else:
            if not axis_on:
                ax.set_xlabel('')
                ax.set_ylabel('')
                ax.set_zlabel('')
                ax.xaxis.set_ticklabels([])
                ax.yaxis.set_ticklabels([])
                ax.zaxis.set_ticklabels([])
            else:
                ax.set_xlabel(a)
                ax.set_ylabel(b)
                ax.set_zlabel(c)
                ax.set_xticklabels(ax.get_zticklabels(), rotation=90, ha='right')

            if not frame_on:
                ax.xaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.yaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.zaxis._axinfo['grid']['color'] = (1, 1, 1, 0)
                ax.xaxis._axinfo['tick']['inward_factor'] = 0
                ax.xaxis._axinfo['tick']['outward_factor'] = 0
                ax.yaxis._axinfo['tick']['inward_factor'] = 0
                ax.yaxis._axinfo['tick']['outward_factor'] = 0
                ax.zaxis._axinfo['tick']['inward_factor'] = 0
                ax.zaxis._axinfo['tick']['outward_factor'] = 0
        count += 1
    for i in range(col+1, n_cols):
        fig.delaxes(axs[row, i])
    fig.tight_layout()


def gene_score_histogram(
    adata,
    score_key: str,
    genes: Optional[List[str]] = None,
    bins: int = 100,
    quantile: Optional[float] = 0.95,
    extra_offset_fraction: float = 0.1,
    anno_min_diff_fraction: float = 0.05,
) -> plt.Figure:
    """
    Draw a histogram of gene scores with percentile line and annotations for specific genes.
    Adapted from Palantir (https://github.com/dpeerlab/Palantir/blob/master/src/palantir/plot.py#L1803)

    Parameters
    ----------
    adata : AnnData
        Annotated data matrix.
    score_key : str
        The key in `adata.var` data frame for the gene score.
    genes : Optional[List[str]], default=None
        List of genes to be annotated. If None, no genes are annotated.
    bins : int, default=100
        The number of bins for the histogram.
    quantile : Optional[float], default=0.95
        Quantile line to draw on the histogram. If None, no line is drawn.
    extra_offset_fraction : float, default=0.1
        Fraction of max height to use as extra offset for annotation.
    anno_min_diff_fraction : float, default=0.05
        Fraction of the range of the scores to be used as minimum difference for annotation.

    Returns
    -------
    fig : matplotlib Figure
        Figure object with the histogram.

    Raises
    ------
    ValueError
        If input parameters are not as expected.
    """
    if score_key not in adata.var.columns:
        raise ValueError(f"Score key {score_key} not found in ad.var columns.")
    scores = adata.var[score_key]

    if genes is not None:
        if not all(gene in scores for gene in genes):
            raise ValueError("All genes must be present in the scores.")

    fig, ax = plt.subplots(figsize=(10, 6))
    n_markers = len(genes) if genes is not None else 0

    heights, bins, _ = ax.hist(scores, bins=bins, zorder=-n_markers - 2)

    if quantile is not None:
        if quantile < 0 or quantile > 1:
            raise ValueError("Quantile should be a float between 0 and 1.")
        ax.vlines(
            np.quantile(scores, quantile),
            0,
            np.max(heights),
            alpha=0.5,
            color="red",
            label=f"{quantile:.0%} percentile",
        )

    ax.legend()
    ax.set_xlabel(f"{score_key} score")
    ax.set_ylabel("# of genes")

    ax.spines[["right", "top"]].set_visible(False)
    plt.locator_params(axis="x", nbins=3)

    if genes is None:
        return fig

    previous_value = -np.inf
    extra_offset = extra_offset_fraction * np.max(heights)
    min_diff = anno_min_diff_fraction * (np.max(bins) - np.min(bins))
    marks = scores[genes].sort_values()
    ranks = scores.rank(ascending=False)
    for k, (highlight_gene, value) in enumerate(marks.items()):
        hl_rank = int(ranks[highlight_gene])
        i = np.searchsorted(bins, value)
        text_offset = -np.inf if value - previous_value > min_diff else previous_value
        previous_value = value
        height = heights[i - 1]
        text_offset = max(text_offset + extra_offset, height + 1.8 * extra_offset)
        txt = ax.annotate(
            f"{highlight_gene} #{hl_rank}",
            (value, height),
            (value, text_offset),
            arrowprops=dict(facecolor="black", width=1, alpha=0.5),
            rotation=90,
            horizontalalignment="center",
            zorder=-k,
        )
        txt.set_path_effects(
            [PathEffects.withStroke(linewidth=2, foreground="w", alpha=0.8)]
        )

    return fig
    

def gene_trend(adata, 
               genes, 
               tkey:str, 
               layer:str='Ms', 
               n_x_grid: int = 100,
               max_iter: int = 2000,
               return_gam_result: bool = False,
               zero_indicator: bool = False,
               sharey: bool = False,
               hide_trend: bool = False,
               hide_cells: bool = False,
               hide_interval: bool = False,
               set_label: bool = True,
               same_plot: bool = False,
               color=None, cmap='plasma', pointsize=1, figsize=None, ncols=5, dpi=100, scatter_kwargs=None, **kwargs):
    """ Plot gene expression or velocity trends along trajectory using Generalized Addictive Model. """
    if tkey not in adata.obs.keys():
        raise ValueError(f'{tkey} not found in adata.obs')
    t = adata.obs[tkey]
    
    if same_plot:
        hide_cells = True
        logging.info('Setting `hide_cells` to True because of plotting trends in the same plot.')
    
    if isinstance(genes, str):
        genes = [genes]
    genes = np.array(genes)
    missing_genes = genes[~np.isin(genes, adata.var_names)]
    if len(missing_genes) > 0:
        print(f'{missing_genes} not found')
    genes = genes[np.isin(genes, adata.var_names)]
    gn = len(genes)
    if gn == 0:
        raise ValueError('genes not found in adata.var_names')
    if gn < ncols:
        ncols = gn

    cell_annot = None
    if color in adata.obs and is_numeric_dtype(adata.obs[color]):
        colors = adata.obs[color].values
    elif color in adata.obs and is_categorical_dtype(adata.obs[color])\
        and color+'_colors' in adata.uns.keys():
        cell_annot = adata.obs[color].cat.categories
        if isinstance(adata.uns[f'{color}_colors'], dict):
            colors = list(adata.uns[f'{color}_colors'].values())
        elif isinstance(adata.uns[f'{color}_colors'], list):
            colors = adata.uns[f'{color}_colors']
        elif isinstance(adata.uns[f'{color}_colors'], np.ndarray):
            colors = adata.uns[f'{color}_colors'].tolist()
        else:
            raise ValueError(f'Unsupported adata.uns[{color}_colors] object')
    else:
        raise ValueError('Currently, color key must be a single string of '
                         'either numerical or categorical available in adata'
                         ' obs, and the colors of categories can be found in'
                         ' adata uns.')

    nrows = -(-gn // ncols)
    fig, axs = plt.subplots(nrows, ncols, squeeze=False,
                            figsize=(6*ncols, 4*(-(-gn // ncols)))
                            if figsize is None else figsize,
                            sharex=True, 
                            sharey=sharey,
                            tight_layout=True,
                            dpi=dpi)

    fig.patch.set_facecolor('white')
    axs = np.reshape(axs, (nrows, ncols))
    logging.info("Plotting trends")

    gam_kwargs = {
        'n_splines': 6,
        'spline_order': 3
    }
    gam_kwargs.update(kwargs)
    gam_results = np.zeros((len(genes), n_x_grid))

    cnt = 0
    for i, gene in tqdm(
        enumerate(genes),
        total=gn,
        desc="Fitting trends using GAM",):
        x = adata[:, gene].layers[layer]
        x = x.A.flatten() if sp.issparse(x) else x.flatten()

        ### GAM fitting
        term  =s(
            0,
            **gam_kwargs)
        gam = LinearGAM(term, max_iter=max_iter, verbose=False).fit(t, x)
        tx = gam.generate_X_grid(term=0, n=n_x_grid)
        row = i // ncols
        col = i % ncols
        ax = axs[row, col] if not same_plot else axs

        if not hide_trend:
            ax.plot(tx[:, 0], gam.predict(tx))
            if not hide_interval:
                ci = gam.confidence_intervals(tx, width=0.95)
                lower_bound, upper_bound = ci[:, 0], ci[:, 1]
                ax.fill_between(tx[:, 0], lower_bound, upper_bound, color='#cabad7', alpha=0.5)
                ax.plot(tx[:, 0], gam.confidence_intervals(tx, width=0.95), c='#cabad7', ls='--')
        if not hide_cells:
            if cell_annot is not None:
                for j in range(len(cell_annot)):
                    filt = adata.obs[color] == cell_annot[j]
                    filt = np.ravel(filt)
                    ax.scatter(t[filt], x[filt], c=colors[j], s=pointsize, alpha=0.5)
            else:
                ax.scatter(t, x, c=colors, s=pointsize, alpha=0.5, cmap=cmap)
        if set_label:
            ax.set_ylabel(gene)

        if zero_indicator:
            ax.axhline(y=0, color='red', linestyle='--')

        if return_gam_result:
            gam_results[i] = gam.predict(tx)
        
        cnt += 1

    if set_label:
        fig.text(0.5, 0.01, tkey, ha='center') 
    for i in range(col+1, ncols):
        fig.delaxes(axs[row, i])
    fig.tight_layout()
    
    if return_gam_result:
        return gam_results
    else:
        return


def response(
    adata,
    pairs_mat: np.ndarray,
    xkey: Optional[str] = 'M_sc',
    ykey: Optional[str] = 'jacobian',
    hide_cells: bool = True,
    annot_key: str = 'celltype',
    downsampling: int = 3,
    log: bool = True,
    drop_zero_cells: bool = True,
    cell_idx: Optional[pd.Index] = None,
    perc: Optional[tuple] = None,
    grid_num: int = 25,
    kde_backend: Literal['fixbdw', 'scipy', 'statsmodels'] = 'statsmodels',
    integral_rule: Literal['trapz', 'simps'] = 'trapz',
    plot_integration_uncertainty: bool = False,
    n_row: int = 1,
    n_col: Optional[int] = None,
    cmap: Union[str, Colormap, None] = None,
    curve_style: str = "c-",
    zero_indicator: bool = True,
    hide_mean: bool = False,
    hide_trend: bool = False,
    figsize: Tuple[float, float] = (6, 4),
    show: bool = True,
    return_integration: bool = False,
    **kwargs,
    ):
    """ Fiting response curve using GAM """
    from scipy import integrate
    from matplotlib.ticker import MaxNLocator
    from mpl_toolkits.axes_grid1.inset_locator import inset_axes
    try:
        from dynamo.vectorfield.utils import get_jacobian
    except ImportError:
        raise ImportError(
            "If you want to show jacobian analysis in plotting function, you need to install `dynamo` "
            "package via `pip install dynamo-release` see more details at https://dynamo-release.readthedocs.io/en/latest/,")
    
    if not set([xkey, ykey]) <= set(adata.layers.keys()).union(set(["jacobian"])):
        raise ValueError(
            f"adata.layers doesn't have {xkey, ykey} layers. Please specify the correct layers or "
            "perform relevant preprocessing and vector field analyses first."
        )
    
    if integral_rule == 'trapz':
        rule = integrate.cumulative_trapezoid
    elif integral_rule == 'simps': 
        rule = integrate.simpson
    else:
        raise ValueError("Integral rule only support `trapz` and `simps` implemented by scipy.") 

    if cmap is None:
        cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
            "response", ["#000000", "#000000", "#000000", "#800080", "#FF0000", "#FFFF00"]
        )
    inset_dict = {
        "width": "5%",  # width = 5% of parent_bbox width
        "height": "50%",  # height : 50%
        "loc": "lower left",
        "bbox_to_anchor": (1.0125, 0.0, 1, 1),
        "borderpad": 0,
    }

    if not hide_cells:
        if annot_key in adata.obs and is_categorical_dtype(adata.obs[annot_key]) \
                and annot_key+'_colors' in adata.uns.keys():
            cell_annot = adata.obs[annot_key].cat.categories
            if isinstance(adata.uns[f'{annot_key}_colors'], dict):
                colors = list(adata.uns[f'{annot_key}_colors'].values())
            else:
                colors = adata.uns[f'{annot_key}_colors']

    all_genes_in_pair = np.unique(pairs_mat)
    if not (set(all_genes_in_pair) <= set(adata.var_names)):
        raise ValueError(
            "adata doesn't include all genes in gene_pairs_mat. Make sure all genes are included in adata.var_names."
        )

    flat_res = pd.DataFrame(columns=["x", "y", "den", "type"])
    xy = pd.DataFrame()
    # extract information from dynamo output in adata object
    id = 0
    for _, gene_pairs in enumerate(pairs_mat):
        f_ini_ind = (grid_num**2) * id

        gene_pair_name = gene_pairs[0] + "->" + gene_pairs[1]

        if xkey.startswith("jacobian"):
            J_df = get_jacobian(
                adata,
                gene_pairs[0],
                gene_pairs[1],
            )
            jkey = gene_pairs[0] + "->" + gene_pairs[1] + "_jacobian"
            x = flatten(J_df[jkey])
        else:
            x = flatten(adata[:, gene_pairs[0]].layers[xkey])

        if ykey.startswith("jacobian"):
            J_df = get_jacobian(
                adata,
                gene_pairs[0],
                gene_pairs[1],
            )
            jkey = gene_pairs[0] + "->" + gene_pairs[1] + "_jacobian"
            y_ori = flatten(J_df[jkey])
        else:
            y_ori = flatten(adata[:, gene_pairs[1]].layers[ykey])

        if drop_zero_cells:
            finite = np.isfinite(x + y_ori)
            nonzero = np.abs(x) + np.abs(y_ori) > 0
            valid_ids = np.logical_and(finite, nonzero)
        else:
            valid_ids = np.isfinite(x + y_ori)

        if cell_idx is not None:
            # subset cells for cell type-specific visualization
            subset_idx = np.zeros(adata.n_obs)
            idx = adata.obs_names.get_indexer(cell_idx)
            subset_idx[idx] = True
            valid_ids = np.logical_and(valid_ids, subset_idx)
        
        if perc is not None:
            # filter out outliers
            lb = np.percentile(x, perc[0])
            ub = np.percentile(x, perc[1])
            valid_ids = np.logical_and(valid_ids, np.logical_and(x>lb, x<ub))

        x, y_ori = x[valid_ids], y_ori[valid_ids]

        if log:
            x, y_ori = x if sum(x < 0) else np.log(np.array(x) + 1), y_ori if sum(y_ori) < 0 else np.log(
                np.array(y_ori) + 1)

        y = y_ori

        # den_res[0, 0] is at the lower bottom; dens[1, 4]: is the 2nd on x-axis and 5th on y-axis
        x_meshgrid, y_meshgrid, den_res = kde2d(
            x, y, n=[grid_num, grid_num], lims=[min(x), max(x), min(y), max(y)], 
            backend=kde_backend
        )
        den_res = np.array(den_res)

        den_x = np.sum(den_res, axis=1)  # condition on each input x, sum over y

        for i in range(len(x_meshgrid)):
            tmp = den_res[i] / den_x[i]  # condition on each input x, normalize over y
            tmp = den_res[i]
            max_val = max(tmp)
            min_val = min(tmp)

            rescaled_val = (tmp - min_val) / (max_val - min_val)
            res_row = pd.DataFrame(
                {
                    "x": x_meshgrid[i],
                    "y": y_meshgrid,
                    "den": rescaled_val,
                    "type": gene_pair_name,
                },
                index=[i * len(x_meshgrid) + np.arange(len(y_meshgrid)) + f_ini_ind],
            )

            flat_res = pd.concat([flat_res, res_row])

        cur_data = pd.DataFrame({"x": x, "y": y, "type": gene_pair_name})
        xy = pd.concat([xy, cur_data], axis=0)
        id = id + 1

    gene_pairs_num = len(flat_res.type.unique())

    # plot jacobian results and fitting curve
    n_col = -(-gene_pairs_num // n_row) if n_col is None else n_col

    if n_row * n_col < gene_pairs_num:
        raise ValueError("The number of row or column specified is less than the gene pairs")
    figsize = (figsize[0] * n_col, figsize[1] * n_row) if figsize is not None else (4 * n_col, 4 * n_row)
    fig, axes = plt.subplots(n_row, n_col, figsize=figsize, sharex=False, sharey=False, squeeze=False)

    def scale_func(x, X, grid_num):
        return grid_num * (x - np.min(X)) / (np.max(X) - np.min(X))

    fit_integration = pd.DataFrame(columns=["x", "y", "mean", "ci_lower", "ci_upper", "regulators", "effectors"])

    for x, flat_res_type in enumerate(flat_res.type.unique()):
        gene_pairs = flat_res_type.split("->")

        flat_res_subset = flat_res[flat_res["type"] == flat_res_type]
        xy_subset = xy[xy["type"] == flat_res_type]

        x_val, y_val = flat_res_subset["x"], flat_res_subset["y"]

        i, j = x % n_row, x // n_row

        values = flat_res_subset["den"].values.reshape(grid_num, grid_num).T

        axins = inset_axes(axes[i, j], bbox_transform=axes[i, j].transAxes, **inset_dict)

        ext_lim = (min(x_val), max(x_val), min(y_val), max(y_val))
        im = axes[i, j].imshow(
            values,
            interpolation="mitchell",
            origin="lower",
            cmap=cmap)
        
        cb = fig.colorbar(im, cax=axins)
        cb.set_alpha(1)
        cb.draw_all()
        cb.locator = MaxNLocator(nbins=3, integer=False)
        cb.update_ticks()

        # closest_x_ind = np.array([np.searchsorted(x_meshgrid, i) for i in xy_subset["x"].values])
        # closest_y_ind = np.array([np.searchsorted(y_meshgrid, i) for i in xy_subset["y"].values])
        # valid_ids = np.logical_and(closest_x_ind < grid_num, closest_y_ind < grid_num)
        # axes[i, j].scatter(closest_x_ind[valid_ids], closest_y_ind[valid_ids], color="gray", alpha=0.1, s=1)

        if xkey.startswith("jacobian"):
            axes[i, j].set_xlabel(r"$\partial f_{%s} / {\partial x_{%s}$" % (gene_pairs[1], gene_pairs[0]))
        else:
            axes[i, j].set_xlabel(gene_pairs[0] + rf" (${xkey}$)")
        if ykey.startswith("jacobian"):
            axes[i, j].set_ylabel(r"$\partial f_{%s} / \partial x_{%s}$" % (gene_pairs[1], gene_pairs[0]))
            axes[i, j].title.set_text(r"$\rho(\partial f_{%s} / \partial x_{%s})$" % (gene_pairs[1], gene_pairs[0]))
        else:
            axes[i, j].set_ylabel(gene_pairs[1] + rf" (${ykey}$)")
            axes[i, j].title.set_text(rf"$\rho_{{{gene_pairs[1]}}}$ (${ykey}$)")

        xlabels = list(np.linspace(ext_lim[0], ext_lim[1], 5))
        ylabels = list(np.linspace(ext_lim[2], ext_lim[3], 5))

        # zero indicator
        if zero_indicator:
            axes[i, j].plot(
                scale_func([np.min(xlabels), np.max(xlabels)], xlabels, grid_num),
                scale_func(np.zeros(2), ylabels, grid_num),
                'w--',
                linewidth=2.0)

        # curve fiting using pygam
        gam_kwargs = {
            'n_splines': 6,
            'spline_order': 3
        }
        gam_kwargs.update(kwargs)
        if ykey.startswith("jacobian"):
            logging.info("Fitting response curve using GAM...")
            x_grid, y_mean, y_sigm = kde2d_to_mean_and_sigma(
                np.array(x_val), np.array(y_val), flat_res_subset["den"].values)
            y_mean[np.isnan(y_mean)] = 0
            term  =s(
                0,
                **gam_kwargs)
            # w = (1/y_sigm - (1/y_sigm).min()) / ((1/y_sigm).max()-(1/y_sigm).min())
            gam = LinearGAM(term, max_iter=1000, verbose=False).fit(x_grid, y_mean)
            if not hide_mean:
                axes[i, j].plot(
                    scale_func(x_grid, xlabels, grid_num), 
                    scale_func(y_mean, ylabels, grid_num), 
                    "c*"
                )
            if not hide_trend:
                axes[i, j].plot(
                    scale_func(x_grid, xlabels, grid_num), 
                    scale_func(gam.predict(x_grid), ylabels, grid_num), 
                    curve_style
                )

            # Integrate the fitted curve using the integration rule
            int_grid = np.linspace(x_grid.min(), x_grid.max(), 100)
            prediction = gam.predict(int_grid)
            confidence_intervals = gam.confidence_intervals(int_grid, width=0.95)
            integral = rule(prediction, int_grid, initial=0)
            mean_integral = integral
            ci_lower = np.zeros_like(integral)
            ci_upper = np.zeros_like(integral)

            # Perform Monte Carlo simulations to calculate confident interval
            if plot_integration_uncertainty: 
                logging.info('Simulation integral interval using Monte Carlo method...')
                n_simulations = 1000
                simulated_integrals = []
                for _ in range(n_simulations):
                    # Simulate the fitted values within their confidence intervals
                    simulated_values = np.random.uniform(confidence_intervals[:, 0], confidence_intervals[:, 1])
                    # Integrate the simulated values
                    simulated_integral = rule(simulated_values, int_grid, initial=0)
                    simulated_integrals.append(simulated_integral)
                simulated_integrals = np.array(simulated_integrals)
                mean_integral = np.mean(simulated_integrals, axis=0)
                ci_lower = np.percentile(simulated_integrals, 2.5, axis=0)
                ci_upper = np.percentile(simulated_integrals, 97.5, axis=0)

            tmp_integration = pd.DataFrame({
                "x": int_grid, "y": integral, "mean": mean_integral, "ci_lower": ci_lower, "ci_upper": ci_upper, 
                "regulators": gene_pairs[0], "effectors": gene_pairs[1]})
            fit_integration = pd.concat([fit_integration, tmp_integration])

        if not hide_cells:
            y_for_cells = y_meshgrid[-2]
            for type_i in range(len(cell_annot)):
                filt = adata[valid_ids].obs[annot_key] == cell_annot[type_i]
                filt = np.ravel(filt)
                tmp_x = xy_subset["x"].values[filt] # NOTE: xy already subset by valid_ids
                tmp_y = np.full_like(tmp_x, y_for_cells)
                axes[i, j].scatter(scale_func(tmp_x[::downsampling], xlabels, grid_num), tmp_y[::downsampling], s=12, color=colors[type_i], alpha=0.8)

        # set the x/y ticks
        inds = np.linspace(0, grid_num - 1, 5, endpoint=True)
        axes[i, j].set_xticks(inds)
        axes[i, j].set_yticks(inds)

        if ext_lim[1] < 1e-3:
            xlabels = ["{:.3e}".format(i) for i in xlabels]
        else:
            xlabels = [np.round(i, 2) for i in xlabels]
        if ext_lim[3] < 1e-3:
            ylabels = ["{:.3e}".format(i) for i in ylabels]
        else:
            ylabels = [np.round(i, 2) for i in ylabels]

        if ext_lim[1] < 1e-3:
            axes[i, j].set_xticklabels(xlabels, rotation=30, ha="right")
        else:
            axes[i, j].set_xticklabels(xlabels)

        axes[i, j].set_yticklabels(ylabels)

    plt.subplots_adjust(left=0.1, right=1, top=0.80, bottom=0.1, wspace=0.1)
    plt.tight_layout()
    if show:
        plt.show()
    else:
        return axes
    
    if return_integration:
        return fit_integration


import os
import logging
import warnings

import numpy as np
from tqdm import tqdm
import scipy.sparse as sp
from scipy.optimize import minimize
from sklearn.decomposition import PCA
from joblib import Parallel, delayed
import matplotlib.pyplot as plt

from .tangent_space import corr_kernel, cos_corr, density_corrected_transition_matrix, _estimate_dt


def regression_phi(
    i: int, 
    X: np.ndarray,
    V: np.ndarray,
    C: np.ndarray,
    nbrs: list,
    a: float = 1.0,
    b: float = 0.0,
    r: float = 1.0,
    loss_func: str = "linear",
    norm_dist: bool = False,
):
    """
    Compute the regression coefficients (phi) for cell i in the tangent space.
    
    Parameters:
        i (int): The index of the cell to process.
        X (np.ndarray): The coordinate matrix (cells x genes).
        V (np.ndarray): The velocity matrix (cells x genes).
        C (np.ndarray): The correlation (or transition) matrix (cells x ?).
        nbrs (list): List of neighbor indices for each cell.
        a (float): Weight for the reconstruction error term.
        b (float): Weight for the cosine similarity term.
        r (float): Weight for the regularization term.
        loss_func (str): Loss function type ('linear' or 'log').
        norm_dist (bool): If True, normalize the difference vectors.
    
    Returns:
        tuple: The cell index and the optimized weight vector (phi) for cell i.
    """
    x, v, c, idx = X[i], V[i], C[i], nbrs[i]
    c = c[idx]

    # normalized differences
    D = X[idx] - x
    if norm_dist:
        dist = np.linalg.norm(D, axis=1)
        dist[dist == 0] = 1
        D /= dist[:, None]

    # co-optimization
    c_norm = np.linalg.norm(c)

    def func(w):
        v_ = w @ D

        # cosine similarity between w and c
        if b == 0:
            sim = 0
        else:
            cw = c_norm * np.linalg.norm(w)
            if cw > 0:
                sim = c.dot(w) / cw
            else:
                sim = 0

        # reconstruction error between v_ and v
        rec = v_ - v
        rec = rec.dot(rec)
        if loss_func is None or loss_func == "linear":
            rec = rec
        elif loss_func == "log":
            rec = np.log(rec)
        else:
            raise NotImplementedError(
                f"The function {loss_func} is not supported. Choose either `linear` or `log`."
            )

        # regularization
        reg = 0 if r == 0 else w.dot(w)

        ret = a * rec - b * sim + r * reg
        return ret

    def fjac(w):
        v_ = w @ D

        # reconstruction error
        jac_con = 2 * a * D @ (v_ - v)

        if loss_func is None or loss_func == "linear":
            jac_con = jac_con
        elif loss_func == "log":
            jac_con = jac_con / (v_ - v).dot(v_ - v)

        # cosine similarity
        w_norm = np.linalg.norm(w)
        if w_norm == 0 or b == 0:
            jac_sim = 0
        else:
            jac_sim = b * (c / (w_norm * c_norm) - w.dot(c) / (w_norm**3 * c_norm) * w)

        # regularization
        if r == 0:
            jac_reg = 0
        else:
            jac_reg = 2 * r * w

        return jac_con - jac_sim + jac_reg

    res = minimize(func, x0=C[i, idx], jac=fjac)
    return i, res["x"]

def tangent_space_projection(
    X: np.ndarray,
    V: np.ndarray,
    C: np.ndarray,
    nbrs: list,
    a: float = 1.0,
    b: float = 0.0,
    r: float = 1.0,
    loss_func: str = "linear",
    n_jobs: int = None,
):
    """
    The function generates a graph based on the velocity data by minimizing the loss function:
                    L(w_i) = a |v_ - v|^2 - b cos(u, v_) + lambda * \sum_j |w_ij|^2
    where v_ = \sum_j w_ij*d_ij. The flow from i- to j-th node is returned as the basis matrix phi[i, j],

    Arguments
    ---------
        X: :class:`~numpy.ndarray`
            The coordinates of cells in the expression space.
        V: :class:`~numpy.ndarray`
            The velocity vectors in the expression space.
        C: :class:`~numpy.ndarray`
            The transition matrix of cells based on the correlation/cosine kernel.
        nbrs: list
            List of neighbor indices for each cell.
        a: float (default 1.0)
            The weight for preserving the velocity length.
        b: float (default 1.0)
            The weight for the cosine similarity.
        r: float (default 1.0)
            The weight for the regularization.
        n_jobs: `int` (default: available threads)
            Number of parallel jobs.

    Returns
    -------
        E: :class:`~numpy.ndarray`
            The phi matrix.
    """
    if (n_jobs is None or not isinstance(n_jobs, int) or n_jobs < 0 or
            n_jobs > os.cpu_count()):
        n_jobs = os.cpu_count()
    if isinstance(n_jobs, int):
        logging.info(f'running {n_jobs} jobs in parallel')

    vgenes = np.ones(V.shape[-1], dtype=bool)
    vgenes &= ~np.isnan(V.sum(0))
    V = V[:, vgenes]
    X = X[:, vgenes]

    E = np.zeros((X.shape[0], X.shape[0]))

    res = Parallel(n_jobs=n_jobs, backend='loky')(
        delayed(regression_phi)(
            i, 
            X,
            V,
            C,
            nbrs,
            a,
            b,
            r,
            loss_func,
        )
        for i in tqdm(
        range(X.shape[0]),
        total=X.shape[0],
        desc="Learning Phi in tangent space projection.",
    ))

    for i, res_x in res:
        E[i][nbrs[i]] = res_x
    
    return E


class GraphVelo():
    """
    GraphVelo encapsulates the workflow for learning a manifold-constrained velocity projection.
    
    It supports computing a low-dimensional representation of the gene expression space,
    estimating a velocity graph (phi coefficients), and transform velocity vectors to different basis.
    """

    def __init__(
        self,
        adata, 
        xkey='Ms', 
        vkey='velocity', 
        X_data=None,
        V_data=None,
        gene_subset=None,
        approx=True,
        n_pcs=30,
        mo=False,):
        """
        Initialize the GraphVelo object.
        
        Parameters:
            adata: AnnData object containing the expression and velocity data.
            xkey (str): Key in adata.layers for the expression data.
            vkey (str): Key in adata.layers for the velocity data.
            X_data, V_data: Optionally provided expression and velocity matrices.
            gene_subset: Optionally, a subset of genes to use.
            approx (bool): If True, perform an approximate projection using PCA.
            n_pcs (int): Number of principal components for dimensionality reduction.
            mo (bool): Flag indicating if multi-omic data is used (affects neighbor extraction).
        """

        if X_data is not None and V_data is not None:
            X = np.array(X_data.A if sp.issparse(X_data)
            else X_data)
            V = np.array(V_data.A if sp.issparse(V_data)
            else V_data)
        else:
            X_org = np.array(
                adata.layers[xkey].A
                if sp.issparse(adata.layers[xkey])
                else adata.layers[xkey]
            )
            V_org = np.array(
                adata.layers[vkey].A
                if sp.issparse(adata.layers[vkey])
                else adata.layers[vkey]
            )
            
            subset = np.ones(adata.n_vars, bool)
            if gene_subset is not None:
                var_names_subset = adata.var_names.isin(gene_subset)
                subset &= var_names_subset if len(var_names_subset) > 0 else gene_subset
            else: 
                gene_subset = adata.var_names
            X = X_org[:, subset]
            V = V_org[:, subset]

            nans = np.isnan(np.sum(V, axis=0))
            logging.info(f"{nans.sum()} genes are removed because of nan velocity values.")
            if np.any(nans):
                X = X[:, ~nans]
                V = V[:, ~nans]
        self.approx = False

        if "neighbors" not in adata.uns.keys():
            # Check construction knn in reduced space.
            if mo:
                if 'WNN' not in adata.uns.keys():
                    logging.error("`WNN` not in adata.uns")
                nbrs_idx = adata.uns['WNN']['indices']
            else:
                raise ValueError("Please run dyn.tl.neighbors first.")
        else:
            nbrs_idx = adata.uns["neighbors"]['indices']
        self.nbrs_idx = nbrs_idx
        
        if approx:
            self.approx = True
            dt = _estimate_dt(X, V, nbrs_idx)
            X_plus_V = X + V * dt 
            X_plus_V[X_plus_V < 0] = 0 
            X = np.log1p(X)
            X_plus_V = np.log1p(X_plus_V)
            pca = PCA(n_components=n_pcs, svd_solver='arpack', random_state=0)
            pca_fit = pca.fit(X)
            X_pca = pca_fit.transform(X)
            Y_pca = pca_fit.transform(X_plus_V)
            V_pca = (Y_pca - X_pca) / dt
            self.X = X_pca
            self.V = V_pca
        else: 
            self.X = X
            self.V = V
        self.params = {'xkey': xkey,
                       'vkey': vkey,
                       'gene_subset': list(gene_subset),
                       'approx': approx,
                       'n_pcs': n_pcs}

    def train(self, a=1, b=10, r=1, loss_func=None, transition_matrix=None, softmax_adjusted=False):
        """
        Train the GraphVelo model by learning the phi coefficients in tangent space.
        
        Parameters:
            a, b, r: Weights for the loss function components.
            loss_func (str): Loss function type; defaults to 'linear' if approx is True, else 'log'.
            transition_matrix: Optionally provided transition matrix; if not provided, it is computed.
            softmax_adjusted (bool): Flag for adjusting the softmax in the correlation kernel.
        """

        if loss_func is None:
            loss_func = 'linear' if self.approx else 'log'
        train_params = {'a': a, 'b': b, 'r': r, 'loss_func': loss_func, 'softmax_adjusted': softmax_adjusted}
        if transition_matrix is None:
            P = corr_kernel(self.X, self.V, self.nbrs_idx, corr_func=cos_corr, softmax_adjusted=softmax_adjusted)
            P_dc = density_corrected_transition_matrix(P).A
        else:
            P_dc = transition_matrix
        T = tangent_space_projection(self.X, self.V, P_dc, self.nbrs_idx, a=a, b=b, r=r, loss_func=loss_func)
        self.T = sp.csr_matrix(T)
        self.params.update(train_params)
    
    def project_velocity(self, X_embedding, T=None) -> np.ndarray:
        """
        Project the velocity vectors onto a low-dimensional embedding.
        
        Parameters:
            X_embedding (np.ndarray): The low-dimensional embedding coordinates.
            T (optional): An external phi basis. If None, uses the learned phi coefficients.
        
        Returns:
            np.ndarray: The projected velocity vectors in the embedding space.
        """

        if T is None:
            T = self.T
        else:
            logging.warning('You are projecting the velocity vectors with an external `phi` basis.')
        n = T.shape[0]
        delta_X = np.zeros((n, X_embedding.shape[1]))

        sparse_emb = False
        if sp.issparse(X_embedding):
            X_embedding = X_embedding.A
            sparse_emb = True

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            for i in tqdm(
                range(n),
                total=n,
                desc="projecting velocity vector to low dimensional embedding",
            ):
                idx = T[i].indices
                diff_emb = X_embedding[idx] - X_embedding[i, None]
                if np.isnan(diff_emb).sum() != 0:
                    diff_emb[np.isnan(diff_emb)] = 0
                T_i = T[i].data
                delta_X[i] = T_i.dot(diff_emb)

        return sp.csr_matrix(delta_X) if sparse_emb else delta_X

    def plot_phi_dist(self):
        """
        Plot the distribution of the learned phi coefficients.
        
        This uses seaborn for visualization.
        """
        try:
            import seaborn as sns
        except ImportError:
            raise ImportError(
                "You need to install `seaborn` for `phi` distribution visualization.")

        T = self.T.A
        sns.distplot(T[T>0])
        plt.show()

    def write_to_adata(self, adata, key=None):
        """
        Write the learned phi coefficients (velocity projection basis) to the AnnData object.
        
        Parameters:
            adata: AnnData object where the phi matrix should be stored.
            key: The key under which the phi parameters will be saved in adata.
                 Defaults to 'gv' if not provided.
        """
        key = 'gv' if key is None else key
        adata.uns[f"{key}_params"] = {
            **adata.uns.get(f"{key}_params", {}),
            **self.params,
        }
        adata.obsp[key] = self.T.copy()


import math
import numpy as np
import pandas as pd

from sklearn.cluster import DBSCAN
from scipy.stats import gaussian_kde


def dnorm(x, u=0, sig=1):
    """
    Generate the gaussian kernel function.
    """
    return np.exp(-((x - u) ** 2) / (2 * sig ** 2)) / (math.sqrt(2 * math.pi) * sig)

def bandwidth_nrd(x):
    """
    Determine the prefered bandwidth h (smooth factor) for kernel density estimation
    """
    
    x = pd.Series(x)
    h = (x.quantile([0.75]).values - x.quantile([0.25]).values) / 1.34

    return 4 * 1.06 * min(math.sqrt(np.var(x, ddof=1)), h) * (len(x) ** (-1 / 5))

def rep(x, length):
    len_x = len(x)
    n = int(length / len_x)
    r = length % len_x
    re = []
    for i in range(0, n):
        re = re + x
    for i in range(0, r):
        re = re + [x[i]]
    return re

def kde2d_scipy(x, y, n=25, lims=None):
    """2D Kernel Density Estimation using scipy with adaptive bandwidth."""
    nx = len(x)
    if not lims:
        lims = [min(x), max(x), min(y), max(y)]
    if len(y) != nx:
        raise ValueError("data vectors must be the same length")
    if not np.isfinite(x).all() or not np.isfinite(y).all() or not np.isfinite(lims).all():
        raise ValueError("data and limits must be finite")

    n = rep(n, length=2) if isinstance(n, list) else rep([n], length=2)
    gx = np.linspace(lims[0], lims[1], n[0])
    gy = np.linspace(lims[2], lims[3], n[1])

    # Estimate the marginal density P(X)
    kde_x = gaussian_kde(x, bw_method='scott')
    p_x = kde_x(gx)

    # Estimate the joint density P(X, Y)
    kde_xy = gaussian_kde([x, y], bw_method='scott')
    positions = np.vstack([np.repeat(gx, len(gy)), np.tile(gy, len(gx))])
    p_xy = kde_xy(positions).reshape(len(gx), len(gy))

    return gx, gy, p_xy

def kde2d_statsmodels(x, y, n=25, lims=None):
    """2D Kernel Density Estimation using statsmodels with adaptive bandwidth."""
    try:
        from statsmodels.nonparametric.kernel_density import KDEMultivariate
    except:
        raise ImportError("If you want to use `statsmodels` as KDE backend, you need to"
                           "install `statsmodels` first")
    nx = len(x)
    if not lims:
        lims = [min(x), max(x), min(y), max(y)]
    if len(y) != nx:
        raise ValueError("data vectors must be the same length")
    if not np.isfinite(x).all() or not np.isfinite(y).all() or not np.isfinite(lims).all():
        raise ValueError("data and limits must be finite")

    n = rep(n, length=2) if isinstance(n, list) else rep([n], length=2)
    gx = np.linspace(lims[0], lims[1], n[0])
    gy = np.linspace(lims[2], lims[3], n[1])

    # Estimate the joint density P(X, Y) using statsmodels
    kde_xy = KDEMultivariate(data=[x, y], var_type='cc', bw='normal_reference')
    positions = np.vstack([np.repeat(gx, len(gy)), np.tile(gy, len(gx))]).T
    p_xy = kde_xy.pdf(positions).reshape(len(gx), len(gy))

    return gx, gy, p_xy

def kde2d_fixbdw(x, y, n=25, lims=None):
    nx = len(x)
    if not lims:
        lims = [min(x), max(x), min(y), max(y)]
    if len(y) != nx:
        raise Exception("data vectors must be the same length")
    elif (False in np.isfinite(x)) or (False in np.isfinite(y)):
        raise Exception("missing or infinite values in the data are not allowed")
    elif False in np.isfinite(lims):
        raise Exception("only finite values are allowed in 'lims'")
    else:
        n = rep(n, length=2) if isinstance(n, list) else rep([n], length=2)
        gx = np.linspace(lims[0], lims[1], n[0])
        gy = np.linspace(lims[2], lims[3], n[1])
        
        h = [bandwidth_nrd(x), bandwidth_nrd(y)]
        if 0 in h:
            max_vec = [max(x), max(y)]
            h[h == 0] = max_vec[h == 0] / n
        h = [item[0] if isinstance(item, np.ndarray) else item for item in h]
        h = np.array(rep(h, length=2))
            
        if h[0] <= 0 or h[1] <= 0:
            raise Exception("bandwidths must be strictly positive")
        else:
            h /= 4
            ax = pd.DataFrame()
            ay = pd.DataFrame()
            for i in range(len(x)):
                ax[i] = (gx - x[i]) / h[0]
            for i in range(len(y)):
                ay[i] = (gy - y[i]) / h[1]
            
            z = (np.matrix(dnorm(ax)) * np.matrix(dnorm(ay).T)) / (nx * h[0] * h[1])
            
    return gx, gy, z


def kde2d(x, y, n=25, lims=None, backend='scipy'):
    if backend == 'scipy':
        return kde2d_scipy(x, y, n, lims)
    elif backend == 'statsmodels':
        return kde2d_statsmodels(x, y, n, lims)
    elif backend == 'fixbdw':
        return kde2d_fixbdw(x, y, n, lims)
    else:
        raise ValueError("Backend not recognized. Choose 'scipy', 'statsmodels' or 'fixbdw'.")


def kde3d(x, y, z, h=None, n=25, lims=None):
    """
    Three-dimensional kernel density estimation with an axis-aligned
    bivariate normal kernel, evaluated on a cubic grid.

    Arguments
    ---------
        x:  `List`
            x coordinate of data
        y:  `List`
            y coordinate of data
        z:  'List'
            z coordinate of data
        
        h:  `List` (Default: None)
            vector of bandwidths for :math:`x`, :math:`y` directions and :math:'z' directions.  
            Defaults to normal reference bandwidth (see `bandwidth.nrd`). A scalar value will 
            be taken to apply to each directions.
        n: `int` (Default: 25)
            Number of grid points in each direction.  Can be scalar or a length-3 integer list.
        lims: `List` (Default: None)
            The limits of the cubic covered by the grid .

    Returns
    ---------
        A list of three components
        gx, gy, gz: `List`
            The x,y and z coordinates of the grid points, lists of length `n`.
        z:  `List`
            An :math:`n[1]` by :math:`n[2]` by :math: 'n[3]' matrix of the estimated density: first dimension corresponds 
            to the value of :math:`x`, second dimension corresponds to the value of :math:`y`, third dimension corresponds
            to the value of :math:'z'.
    """
    nx = len(x)
    ny = len(y)
    if not lims:
        lims = [min(x), max(x), min(y), max(y), min(z), max(z)]
    if not nx == ny == len(z):
        raise Exception("data vectors must be the same length")
    
    elif (False in np.isfinite(x)) or (False in np.isfinite(y)):
        raise Exception("missing or infinite values in the data are not allowed")
    elif False in np.isfinite(lims):
        raise Exception("only finite values are allowed in 'lims'")
    else:
        n = rep(n, length=3) if isinstance(n, list) else rep([n], length=3)
        gx = np.linspace(lims[0], lims[1], n[0])
        gy = np.linspace(lims[2], lims[3], n[1])
        gz = np.linspace(lims[4], lims[5], n[1])
        if h is None:
            h = [bandwidth_nrd(x), bandwidth_nrd(y), bandwidth_nrd(z)]
        else:
            h = np.array(rep(h, length=3))

        if h[0] <= 0 or h[1] <= 0 or h[2] <= 0:
            raise Exception("bandwidths must be strictly positive")
        else:
            h /= 4
            ax = pd.DataFrame()
            ay = pd.DataFrame()
            az = pd.DataFrame()
            for i in range(len(x)):
                ax[i] = (gx - x[i]) / h[0]
            for i in range(len(y)):
                ay[i] = (gy - y[i]) / h[1]
            for i in range(len(z)):
                az[i] = (gz - z[i]) / h[2]
                
            z = np.einsum('ai, bi, ci -> abc',np.matrix(dnorm(ax)),np.matrix(dnorm(ay)),np.matrix(dnorm(az)))/ (nx* ny * h[0] * h[1]*h[2])
            
    return gx, gy, gz, z

def rescale_density(dens):
    """
    Rescale the preliminary density from the kernel density estimation methods. 
    This function can be applied to (n,n) or (n,n,n) dimensional density matrix.

    Arguments
    ---------
        dens: 'numpy.array'
                numpy.array object with dimension of (n,n) or (n,n,n)

    Returns
    ---------
        rescaled_dens: 'numpy.array'
                numpy.array object with dimension of (n,n) or (n,n,n)
    """
    ndim = dens.ndim
    dens_shape = dens.shape[0]
    dens = np.array(dens)
    
    if ndim == 2:
        
        den_x = np.sum(dens, axis = 1) #condition on each input x, sum over y
        rescaled_dens = np.zeros([dens_shape, dens_shape])
        
        for i in range(dens_shape):
            tmp = dens[i] / den_x[i]  #condition on each input x, normalize over y
            max_val = max(tmp)
            min_val = min(tmp)
            rescaled_val = (tmp - min_val) / (max_val - min_val)
            rescaled_dens[:,i] = rescaled_val
            
    elif ndim == 3:
        
        den_xy = np.sum(dens, axis=2)  # condition on each input x,y, sum over z
        rescaled_dens = np.zeros([dens_shape, dens_shape, dens_shape])
        
        for i in range(dens_shape):
            for j in range(dens_shape): 
                tmp = dens[i,j] / den_xy[i,j]  # condition on each input x, normalize over y
                max_val = max(tmp)
                min_val = min(tmp)
                rescaled_val = (tmp - min_val) / (max_val - min_val)
                rescaled_dens[i,j,:] =  rescaled_val
    
    else:
        print('We have not defined the kde method for more than 3-dimensional data.')
    
    return rescaled_dens

def kde2d_to_mean_and_sigma(gx, gy, dens):
    """
    Use density as weight to smooth data on y-axis. gx, gy are meshgrid matrix, 
    dens should be the density matrix after rescaling.

    Arguments
    ---------
        gx: 'numpy.array'
            numpy.array object, with dimension of (n,n)
        gy: 'numpy.array'
            numpy.array object, with dimension of (n,n)
        dens:'numpy.array'
            numpy.array object, with dimension of (n,n)
            
    Returns
    ---------
        x_grid: 'numpy.array'
            numpy.array object, with dimension of (n,)
        y_mean: 'numpy.array'
            numpy.array object, with dimension of (n,)
        y_sigm: 'numpy.array'
            numpy.array object, with dimension of (n,)      
    """
    
    x_grid = np.unique(gx)
    y_mean = np.zeros(len(x_grid))
    y_sigm = np.zeros(len(x_grid))
    for i, x in enumerate(x_grid):
        mask = gx == x
        den = dens[mask]
        Y_ = gy[mask]
        mean = np.average(Y_, weights = den)
        sigm = np.sqrt(np.average((Y_ - mean) ** 2, weights=den))
        y_mean[i] = mean
        y_sigm[i] = sigm
    return x_grid, y_mean, y_sigm

def kde3d_to_mean_and_sigma(gx, gy, gz, dens):
    
    """
    Use density as weight to smooth data on z-axis. gx, gy, gz are meshgrid matrix, 
    dens should be the density matrix after rescaling.

    Arguments
    ---------
        gx: 'numpy.array'
            numpy.array object, with dimension of (n,n)
        gy: 'numpy.array'
            numpy.array object, with dimension of (n,n)
        gz: 'numpy.array'
            numpy.array object, with dimension of (n,n)
        dens:'numpy.array'
            numpy.array object, with dimension of (n,n,n)
            
    Returns
    ---------
        x_grid: 'numpy.array'
            numpy.array object, with dimension of (n,)
        y_grid: 'numpy.array'
            numpy.array object, with dimension of (n,)
        z_mean: 'numpy.array'
            numpy.array object, with dimension of (n,n)
        z_sigm: 'numpy.array'
            numpy.array object, with dimension of (n,n)      
    """
    
    x_grid = np.unique(gx)
    y_grid = np.unique(gy)
    
    z_mean = np.zeros(len(x_grid)*len(y_grid)).reshape(len(x_grid),len(y_grid))
    z_sigm = np.zeros(len(x_grid)*len(y_grid)).reshape(len(x_grid),len(y_grid))
    
    for i, x in enumerate(x_grid):
        for j, y in enumerate(y_grid):
            
            mask1 = gx == x
            mask2 = gy == y
            mask  = mask1 & mask2
            den = dens[mask]
            z_ = gz[mask]
            mean = np.average(z_, weights=den)
            sigm = np.sqrt(np.average((z_ - mean) ** 2, weights=den))
            z_mean[i,j] = mean
            z_sigm[i,j] = sigm

    return x_grid, y_grid, z_mean, z_sigm


def velocity_kde(adata, regulator, coregulator, effector, axis_layer = 'M_t', drop_zero_cells = True, remove_outlier = True):
    
    """
    Calculate the velocity value of specified effector, 
    then perform the kernel density estimation. Return 
    the grid points' coordinates along each axis and the 
    density at each grid points.
    
    Arguments
    ---------
        adata:       'AnnData object'
        regultor:    `String`
                      Name of the regulator gene.
        
        coregultor:    `String`
                      Name of the regulator gene.

        effector:    'String'
                      Name of the effector gene.

        axis_layer:  'String'
                     Specify the layer.

        drop_zero_cells: 'bool'
                        If True, drop out all the zero expression data points.
            
    Returns
    ---------
        gx: 'numpy.array'
            The grid points' x axis coordinates | Dimension (n,).
            
        gy: 'numpy.array'
            The grid points' y axis coordinates | Dimension (n,).
            
        gz: 'numpy.array'
            The grid points' z axis coordinates | Dimension (n,).
            
        dens:  'numpy.array'
            The densities on each grid point | Dimension (n,n,n).
            
        x_axis: 'numpy.array'
            The regulator's expression array
        
        y_axis: 'numpy.array'
            The coregulator's expression array
            
        z_velocity: 'numpy.array'
            The corresponding velocity value of effector gene


    """
    # Get the effector gene(z)'s index within adata's data frame
    z_gene_index = [i for i, g in enumerate(adata.var_names) if g in effector]

    
    # Get the gene expression profile for regulator(x), coregulator(y) and effector(z)
    x_axis = adata[:, regulator].layers[axis_layer].A.flatten()
    y_axis = adata[:, coregulator].layers[axis_layer].A.flatten()
    z_axis = adata[:, effector].layers[axis_layer].A.flatten()
    
    # Calculate the velocity (model 2)
    v = adata.layers['velocity_hill']
    z_velocity = v[:,z_gene_index]

    # Drop zero expression cells
    if drop_zero_cells:
        finite = np.isfinite(x_axis + y_axis + z_axis)
        nonzero = np.abs(x_axis) + np.abs(y_axis)+np.abs(z_axis) > 0
        valid_ids = np.logical_and(finite, nonzero)
        x_axis = x_axis[valid_ids]
        y_axis = y_axis[valid_ids]
        z_axis = z_axis[valid_ids]
        z_velocity = z_velocity[valid_ids].toarray().flatten()
    
    # Remove outlier 
    if remove_outlier:
        points = np.column_stack((x_axis, y_axis, z_velocity))
        # Perform DBSCAN clustering
        dbscan = DBSCAN(eps= 0.9, min_samples=10) # How to choose appropriate hyperparameters?
        labels = dbscan.fit_predict(points)
        x_axis = points[labels == 0, 0]
        y_axis = points[labels == 0, 1]
        z_velocity = points[labels == 0, 2]
        
        # x_axis_outlier = points[labels == -1, 0]
        # y_axis_outlier = points[labels == -1, 1]
        # z_velocity_outlier = points[labels == -1, 2]

    # Setting appropriate bandwidth for kde 
    bandwidth = [bandwidth_nrd(x_axis), bandwidth_nrd(y_axis), bandwidth_nrd(z_velocity)]
    bandwidth  = [item.flatten()[0] if isinstance(item, np.ndarray) else item for item in bandwidth]
    print(f"bandwith shape:{len(bandwidth)}")
    print(f"bandwith:{bandwidth}")
    
    # Setting limits of grid points for kde
    lims=[min(x_axis), max(x_axis), min(y_axis), max(y_axis), min(z_velocity), max(z_velocity)]
    
    #Applying kde3d function 
    gx, gy, gz, dens=kde3d(x_axis, y_axis, z_velocity, h= bandwidth,lims = lims)
    
    return gx, gy, gz, dens, x_axis, y_axis, z_velocity

def jacobian_kde(adata, regulator, effector, axis_layer = 'M_t', drop_zero_cells =True):
    """
    Calculate the jacobian value of specified regulator and effector,
    then perform the kernel density estimation. Return the grid points' coordinates
    along each axis and the density at each grid points.

    Arguments
    ---------
        regultor:    `String`
                      Name of the regulator gene.

        effector:    'String'
                      Name of the effector gene.

        axis_layer:  'String'
                     Specify the layer.

        drop_zero_cells: 'bool'
                        If True, drop out all the zero expression data points.

    Returns
    -------
        gx: 'numpy.array'
            The grid points' x axis coordinates | Dimension (n,).
            
        gy: 'numpy.array'
            The grid points' y axis coordinates | Dimension (n,).
            
        dens:  'numpy.array'
            The densities on each grid point | Dimension (n,n).
            
        x_axis: 'numpy.array'
            The regulator's expression array
            
        y_jacobian: 'numpy.array'
            The corresponding jacobian value of (/partial effector) / (/partial regulator)

    """
    try:
        from dynamo import dyn
    except ImportError:
        raise ImportError(
            "If you want to show jacobian analysis in plotting function, you need to install `dynamo` "
            "package via `pip install dynamo-release` see more details at https://dynamo-release.readthedocs.io/en/latest/,")
    
    #Get the gene expression profile for regulator(x) and effector(y) 
    x_axis = adata[:, regulator].layers[axis_layer].A.flatten().reshape(-1,1)
    y_axis = adata[:, effector].layers[axis_layer].A.flatten().reshape(-1,1)

    # #Get the cell_type information
    # keys_to_check = ['cell_type', 'celltype', 'dyn_phase', 'cell_cycle_phase']
    # cell_type_key = next((key for key in keys_to_check if key in adata.obs.columns), None)
    # cell_type = adata.obs[cell_type_key]

    # color_mapper = get_mapper()
    # color_series = cell_type.map(color_mapper)
    # cell_colors = np.array(color_series).reshape((-1, 1))
    # cell_types = np.array(cell_type).reshape((-1, 1))

    #Calculate the Jacobian
    dyn.vf.jacobian(adata, regulators = [regulator, effector], effectors=[regulator, effector])

    #Specify the gene index within the jacobiann result
    x_gene_index = [i for i, g in enumerate(adata.uns['jacobian_pca']['regulators']) if g in regulator]
    y_gene_index = [i for i, g in enumerate(adata.uns['jacobian_pca']['effectors']) if g in effector]


    #Get Jacobian data for y axis
    if regulator != effector:
        y_jacobian = adata.uns['jacobian_pca']['jacobian_gene'][y_gene_index, x_gene_index, :].reshape(-1,1)
    else:
        y_jacobian = adata.uns['jacobian_pca']['jacobian_gene'].reshape(-1,1)

    #Drop zero expression cells
    if drop_zero_cells:
        # NOTE: For different regulators, valid_ids maybe different. Here, we check 
        # whether there are universal valid ids, then decide whether we need to calculte 
        # it seperately
        if 'universal_valid_id' in adata.uns:
            valid_ids = adata.uns['universal_valid_id']
        else:
            finite = np.isfinite(x_axis + y_axis)
            nonzero = np.abs(x_axis) > 0
            valid_ids = np.logical_and(finite, nonzero)
        x_axis = x_axis[valid_ids].squeeze()
        y_axis = y_axis[valid_ids].squeeze()
        y_jacobian = y_jacobian[valid_ids].squeeze()
        # cell_colors = cell_colors[valid_ids].squeeze()
        # cell_types = cell_types[valid_ids].squeeze()

    # Setting appropriate bandwidth for kde
    bandwidth = [bandwidth_nrd(x_axis), bandwidth_nrd(y_jacobian)]

    # Setting limits of grid points for kde
    lims=[min(x_axis), max(x_axis), min(y_jacobian), max(y_jacobian)]
    # Applying kde2d function 
   
    gx, gy, dens=kde2d(x_axis, y_jacobian, h = bandwidth,lims = lims)


    return gx, gy, dens, x_axis, y_jacobian

def jacobian_kde_3d(adata, regulator, coregulator, effector, axis_layer = 'M_t', drop_zero_cells =True, remove_outlier = True):
    """
    Calculate the jacobian value of specified regulator and effector,
    then perform the kernel density estimation. Return the grid points' coordinates
    along each axis and the density at each grid points.

    Arguments
    ---------
        regultor:    `String`
                      Name of the regulator gene.

        effector:    'String'
                      Name of the effector gene.

        axis_layer:  'String'
                     Specify the layer.

        drop_zero_cells: 'bool'
                        If True, drop out all the zero expression data points.

    Returns
    -------
        gx: 'numpy.array'
            The grid points' x axis coordinates | Dimension (n,).
            
        gy: 'numpy.array'
            The grid points' y axis coordinates | Dimension (n,).
            
        dens:  'numpy.array'
            The densities on each grid point | Dimension (n,n).
            
        x_axis: 'numpy.array'
            The regulator's expression array
            
        y_jacobian: 'numpy.array'
            The corresponding jacobian value of (/partial effector) / (/partial regulator)

    """
    #Get the gene expression profile for regulator(x), coregulator(y) and effector(z)
    try:
        from dynamo import dyn
    except ImportError:
        raise ImportError(
            "If you want to show jacobian analysis in plotting function, you need to install `dynamo` "
            "package via `pip install dynamo-release` see more details at https://dynamo-release.readthedocs.io/en/latest/,")
    
    x_axis = adata[:, regulator].layers[axis_layer].A.flatten().reshape(-1,1)
    y_axis = adata[:, coregulator].layers[axis_layer].A.flatten().reshape(-1,1)
    z_axis = adata[:, effector].layers[axis_layer].A.flatten().reshape(-1,1)

    #Calculate the Jacobian
    dyn.vf.jacobian(adata, regulators = [regulator, effector], effectors=[regulator, effector])

    #Specify the gene index within the jacobiann result
    x_gene_index = [i for i, g in enumerate(adata.uns['jacobian_pca']['regulators']) if g in regulator]
    z_gene_index = [i for i, g in enumerate(adata.uns['jacobian_pca']['effectors']) if g in effector]

    #Get Jacobian data for y axis
    if regulator != effector:
        z_jacobian = adata.uns['jacobian_pca']['jacobian_gene'][z_gene_index, x_gene_index, :].reshape(-1,1)
    else:
        z_jacobian = adata.uns['jacobian_pca']['jacobian_gene'].reshape(-1,1)

    #Drop zero expression cells
    if drop_zero_cells:
        finite = np.isfinite(x_axis + y_axis + z_axis)
        nonzero = np.abs(x_axis) + np.abs(y_axis) + np.abs(z_axis) > 0
        valid_ids = np.logical_and(finite, nonzero)
        x_axis = x_axis[valid_ids]
        y_axis = y_axis[valid_ids]
        z_axis = z_axis[valid_ids]
        z_jacobian = z_jacobian[valid_ids]
    #Remove Outlier
    if remove_outlier:
        points = np.column_stack((x_axis, y_axis, z_jacobian))
        # Perform DBSCAN clustering
        dbscan = DBSCAN(eps= 0.09, min_samples=5)
        labels = dbscan.fit_predict(points)
        x_axis = points[labels == 0, 0]
        y_axis = points[labels == 0, 1]
        z_jacobian = points[labels == 0, 2]
        
        # x_axis_outlier = points[labels == -1, 0]
        # y_axis_outlier = points[labels == -1, 1]
        # z_jacobian_outlier = points[labels == -1, 2]

    # Setting appropriate bandwidth for kde 
    bandwidth = [bandwidth_nrd(x_axis), bandwidth_nrd(y_axis), bandwidth_nrd(z_jacobian)]

    # Setting limits of grid points for kde
    lims=[min(x_axis), max(x_axis), min(y_axis), max(y_axis), min(z_jacobian), max(z_jacobian)]

    # Use kde3d function to calculate the density 
    gx, gy, gz, dens = kde3d(x_axis, y_axis, z_jacobian, h=bandwidth,lims = lims)

    return gx, gy, gz, dens, x_axis, y_axis, z_jacobian


from typing import Optional, Literal
from pygam import LinearGAM, s

import logging
from tqdm import tqdm
import numpy as np
import pandas as pd
import scipy.sparse as sp
from scipy.optimize import minimize_scalar
from anndata import AnnData

from .utils import flatten
from .kernel_density_smooth import kde2d


def fit_velo_peak(
    adata, 
    genes, 
    tkey: str, 
    layer: str = 'Ms', 
    log_norm: bool = True,
    max_iter: int = 2000,
    **kwargs
):
    """
    Identify the peak velocity (i.e. maximum predicted value) for each gene using GAM.
    
    For each gene in `genes`, a GAM is fit using the independent variable stored in 
    adata.obs[tkey] and the expression values from adata[:, gene].layers[layer]. 
    Then the time (or phase) corresponding to the maximum predicted expression is found 
    by minimizing the negative of the GAM prediction.
    
    Parameters:
        adata (anndata.AnnData): AnnData object containing the data.
        genes (str or list of str): One or more gene names.
        tkey (str): Key in adata.obs for the independent variable (e.g., pseudotime).
        layer (str): Name of the layer to use for expression values (default 'Ms').
        log_norm (bool): Whether to apply log1p normalization to the magnitude.
        max_iter (int): Maximum iterations for GAM fitting.
        **kwargs: Additional keyword arguments for GAM (e.g., n_splines, spline_order).
    
    Returns:
        pd.DataFrame: A DataFrame with gene names as the index and two columns:
                      'phase' for the peak time (or phase) and 'magnitude' for the 
                      predicted expression at the peak (optionally log-normalized).
    """
    
    # Check that tkey is present
    if tkey not in adata.obs.keys():
        raise ValueError(f'{tkey} not found in adata.obs')
    t = adata.obs[tkey].values  # Independent variable
    
    # Process gene list
    if isinstance(genes, str):
        genes = [genes]
    genes = np.array(genes)
    missing_genes = genes[~np.isin(genes, adata.var_names)]
    if len(missing_genes) > 0:
        logging.warning(f'{missing_genes} not found')
    genes = genes[np.isin(genes, adata.var_names)]
    gn = len(genes)
    
    # Set up GAM keyword arguments (default: 7 splines, order 3)
    gam_kwargs = {
        'n_splines': 7,
        'spline_order': 3
    }
    gam_kwargs.update(kwargs)
    
    phase = np.zeros(gn)
    magnitude = np.zeros(gn)
        
    for i, gene in tqdm(enumerate(genes), total=gn, desc="Identify velocity peaks using GAM"):
        x = adata[:, gene].layers[layer]
        x = flatten(x)
        
        gam = LinearGAM(s(0, **gam_kwargs), max_iter=max_iter, verbose=False).fit(t, x)
        def neg_gam(x_val):
            return -gam.predict(np.array([[x_val]]))[0]
        
        res = minimize_scalar(neg_gam, bounds=(t.min(), t.max()), method='bounded')
        peak_t = res.x
        phase[i] = peak_t
        magnitude[i] = gam.predict(np.array([[peak_t]]))[0]
    
    # Optionally apply log normalization to the magnitude， genes with negative peak velo can cause error.
    if log_norm:
        magnitude = np.log1p(magnitude)
    
    df = pd.DataFrame({'phase': phase, 'magnitude': magnitude}, index=genes)
    return df
    

def fit_gene_trend(
    adata, 
    genes, 
    tkey:str, 
    layer:str='Ms', 
    max_iter: int = 2000,
    grid_num: int = 200,
    **kwargs):
    """
    Fit gene expression trends over a given time or pseudotime variable using Generalized
    Additive Models (GAMs) and return the predictions in an AnnData object.

    Parameters:
    -----------
    adata : AnnData
        An AnnData object containing the expression data in adata.layers and the
        time/pseudotime information in adata.obs.
    genes : str or list of str
        Gene name (or list of gene names) for which to fit the trend.
    tkey : str
        Key in adata.obs that contains the time or pseudotime variable.
    layer : str, default 'Ms'
        The layer in adata.layers to extract expression values from.
    max_iter : int, default 2000
        Maximum iterations for fitting the GAM.
    grid_num : int, default 200
        Number of points in the grid over which the trend is predicted.
    **kwargs : dict
        Additional keyword arguments to pass to the GAM fitting routine (e.g., n_splines, spline_order).

    Returns:
    --------
    gdata : AnnData
        An AnnData object where .X is a matrix of shape (number of genes, grid_num) containing
        the predicted trend for each gene, and the .obs_names are set to the corresponding gene names.
    """
    
    if tkey not in adata.obs.keys():
        raise ValueError(f'{tkey} not found in adata.obs')
    t = adata.obs[tkey]

    if isinstance(genes, str):
        genes = [genes]
    genes = np.array(genes)
    missing_genes = genes[~np.isin(genes, adata.var_names)]
    if len(missing_genes) > 0:
        print(f'{missing_genes} not found')
    genes = genes[np.isin(genes, adata.var_names)]
    gn = len(genes)

    gam_kwargs = {
        'n_splines': 7,
        'spline_order': 3
    }
    gam_kwargs.update(kwargs)

    data = np.zeros((gn, grid_num)) # N_genes * N_grid

    for i, gene in tqdm(
        enumerate(genes),
        total=gn,
        desc="Fitting trends using GAM",):
        x = adata[:, gene].layers[layer]
        x = x.A.flatten() if sp.issparse(x) else x.flatten()

        ### GAM fitting
        term  =s(
            0,
            **gam_kwargs)
        gam = LinearGAM(term, max_iter=max_iter, verbose=False).fit(t, x)
        x_lins = np.linspace(t.min(), t.max(), grid_num)
        y_pred = gam.predict(x_lins)
        data[i] = y_pred
    
    gdata = AnnData(data)
    gdata.obs_names = pd.Index(genes)
    return gdata


def fit_response(
    adata: AnnData,
    pairs_mat: np.ndarray,
    xkey: Optional[str] = 'M_sc',
    ykey: Optional[str] = 'jacobian',
    norm: bool = True,
    log: bool = False,
    grid_num: int = 200,
    **kwargs,
):
    """
    Fit response curves for gene pairs and generate a grid of GAM predictions.

    For each gene pair provided in pairs_mat, this function:
      - Retrieves the response values (x) from the specified layer (xkey) for the source gene.
      - Computes the Jacobian (y) for the target gene using dynamo's get_jacobian function.
      - Filters out invalid data points (non-finite or zero sums).
      - Optionally applies log transformation and normalization.
      - Fits a Generalized Additive Model (GAM) to predict the response over a grid.
      - Stores the predicted response curve into a grid (data).

    The function returns an AnnData object where:
      - The .X matrix contains the grid data for each gene pair.
      - The .obs_names are set to the target gene names from each pair.
    """

    try:
        from dynamo.vectorfield.utils import get_jacobian
    except ImportError:
        raise ImportError(
            "If you want to do jacobian analysis related to dynamo, you need to install `dynamo` "
            "package via `pip install dynamo-release` see more details at https://dynamo-release.readthedocs.io/en/latest/,")
    
    if not set([xkey, ykey]) <= set(adata.layers.keys()).union(set(["jacobian"])):
        raise ValueError(
            f"adata.layers doesn't have {xkey, ykey} layers. Please specify the correct layers or "
            "perform relevant preprocessing and vector field analyses first."
        )
    
    all_genes_in_pair = np.unique(pairs_mat)
    if not (set(all_genes_in_pair) <= set(adata.var_names)):
        raise ValueError(
            "adata doesn't include all genes in gene_pairs_mat. Make sure all genes are included in adata.var_names."
        )
    if not ykey.startswith("jacobian"):
        raise KeyError('The ykey should start with `jacobian`.')
    
    genes = []
    xy = pd.DataFrame()
    id = 0
    for _, gene_pairs in enumerate(pairs_mat):
        gene_pair_name = gene_pairs[0] + "->" + gene_pairs[1]
        genes.append(gene_pairs[1])

        x = flatten(adata[:, gene_pairs[0]].layers[xkey])
        J_df = get_jacobian(
            adata,
            gene_pairs[0],
            gene_pairs[1],
        )
        jkey = gene_pairs[0] + "->" + gene_pairs[1] + "_jacobian"
        y_ori = flatten(J_df[jkey])

        finite = np.isfinite(x + y_ori)
        nonzero = np.abs(x) + np.abs(y_ori) > 0
        valid_ids = np.logical_and(finite, nonzero)

        x, y_ori = x[valid_ids], y_ori[valid_ids]

        if log:
            x, y_ori = x if sum(x < 0) else np.log(np.array(x) + 1), y_ori if sum(y_ori) < 0 else np.log(
                np.array(y_ori) + 1)

        if norm:
            y_ori = y_ori / y_ori.max()

        y = y_ori

        cur_data = pd.DataFrame({"x": x, "y": y, "type": gene_pair_name})
        xy = pd.concat([xy, cur_data], axis=0)

        id = id + 1

    data = np.zeros((len(pairs_mat), grid_num)) # N_genes * N_grid

    for gene_idx, res_type in enumerate(xy.type.unique()):
        gene_pairs = res_type.split("->")
        xy_subset = xy[xy["type"]==res_type]
        x_val, y_val = xy_subset["x"], xy_subset["y"]

        gam_kwargs = {
            'n_splines': 12,
            'spline_order': 3
        }
        gam_kwargs.update(kwargs)
        term  =s(
            0,
            **gam_kwargs)
        gam = LinearGAM(term, max_iter=1000, verbose=False).fit(x_val, y_val)
        
        x_lins = np.linspace(x_val.min(), x_val.max(), grid_num)
        y_pred = gam.predict(x_lins)
        data[gene_idx] = y_pred
    
    adata = AnnData(data)
    adata.obs_names = pd.Index(genes)

    return adata


def cluster_response(
    adata: AnnData,
    pairs_mat: np.ndarray,
    xkey: Optional[str] = 'M_sc',
    ykey: Optional[str] = 'jacobian',
    grid_num: int = 25,
    kde_backend: Literal['fixbdw', 'scipy', 'statsmodels'] = 'statsmodels',
):
    """
    Cluster responses for gene pairs by computing 2D density estimates on a grid.

    For each gene pair:
      - Retrieves the response values (x) from the specified xkey layer.
      - Retrieves the Jacobian values (y) using dynamo's get_jacobian.
      - Uses a 2D kernel density estimation (kde2d) to compute density over a grid.
      - Normalizes the density over y for each x (row normalization).
      - Flattens the grid and stores it for clustering analysis.

    The function returns a DataFrame where:
      - Each row corresponds to a gene pair (indexed by "gene_pair_name").
      - The columns are the flattened density values on the grid.
    """
    
    try:
        from dynamo.vectorfield.utils import get_jacobian
    except ImportError:
        raise ImportError(
            "If you want to do jacobian analysis related to dynamo, you need to install `dynamo` "
            "package via `pip install dynamo-release` see more details at https://dynamo-release.readthedocs.io/en/latest/,")
    
    if not set([xkey, ykey]) <= set(adata.layers.keys()).union(set(["jacobian"])):
        raise ValueError(
            f"adata.layers doesn't have {xkey, ykey} layers. Please specify the correct layers or "
            "perform relevant preprocessing and vector field analyses first."
        )
    
    all_genes_in_pair = np.unique(pairs_mat)
    if not (set(all_genes_in_pair) <= set(adata.var_names)):
        raise ValueError(
            "adata doesn't include all genes in gene_pairs_mat. Make sure all genes are included in adata.var_names."
        )
    if not ykey.startswith("jacobian"):
        raise KeyError('The ykey should start with `jacobian`.')
    
    flat_data = np.zeros((len(pairs_mat), grid_num**2))
    data_idx = []
    # extract information from dynamo output in adata object
    for gene_idx, gene_pairs in enumerate(pairs_mat):
        gene_pair_name = gene_pairs[0] + "->" + gene_pairs[1]
        data_idx.append(gene_pair_name)

        x = flatten(adata[:, gene_pairs[0]].layers[xkey])

        J_df = get_jacobian(
            adata,
            gene_pairs[0],
            gene_pairs[1],
        )
        jkey = gene_pairs[0] + "->" + gene_pairs[1] + "_jacobian"
        y_ori = flatten(J_df[jkey])
        y = y_ori

        # den_res[0, 0] is at the lower bottom; dens[1, 4]: is the 2nd on x-axis and 5th on y-axis
        x_meshgrid, y_meshgrid, den_res = kde2d(
            x, y, n=[grid_num, grid_num], lims=[-max(x), max(x), -max(y), max(y)], 
            backend=kde_backend
        )
        den_res = np.array(den_res)
        den_x = np.sum(den_res, axis=1)

        data = np.zeros_like(den_res)

        for i in range(len(x_meshgrid)):
            tmp = den_res[i] / den_x[i]  # condition on each input x, normalize over y
            tmp = den_res[i]
            max_val = max(tmp)
            min_val = min(tmp)

            rescaled_val = (tmp - min_val) / (max_val - min_val)
            data[i] = rescaled_val

        data = flatten(data)
        flat_data[gene_idx] = data

    def scale_func(x, X, grid_num):
        return grid_num * (x - np.min(X)) / (np.max(X) - np.min(X))

    data = pd.DataFrame(flat_data, index=data_idx)

    return data



import glob
import os
import re
import subprocess
from logging import info, warning
from pathlib import Path
from shutil import copy, copytree, rmtree
from tempfile import TemporaryDirectory
from typing import Dict, ForwardRef, List, Union

from git import Repo

CUR_DIR = Path(__file__).parent

NOTEBOOK_BRANCH = "main"
LIVECELLX_NOTEBOOK_PATH_ENV_VAR = "GRAPHVELO_DOWNLOAD_NOTEBOOKS"


def _download_docs_dirs(repo_url: str) -> None:
    def copy_docs_dirs(repo_path: Union[str, Path]) -> None:
        repo_path = Path(repo_path)
        print("repo path:", repo_path)
        for dirname in ["notebooks", "gallery", "_static", "tutorials"]:
            if (repo_path / dirname).is_dir():
                print(">>> copying %s" % dirname)
                rmtree(dirname, ignore_errors=True)  # locally re-cloning
                copytree(repo_path / dirname, dirname)

        # Copy all rsts in docs/source/*.rst
        for file_path in glob.glob(str(repo_path / "docs" / "source" / "*.rst")):
            print("%s copied to source" % file_path)
            copy(file_path, "./")  # dest: source

    def fetch_remote(repo_url: str) -> None:
        info(f"Fetching notebooks from repo `{repo_url}`")
        with TemporaryDirectory() as repo_dir:
            branch = NOTEBOOK_BRANCH
            repo = Repo.clone_from(repo_url, repo_dir, depth=1, branch=branch)
            repo.git.checkout(branch, force=True)
            copy_docs_dirs(repo_dir)

    def fetch_local(repo_path: Union[str, Path]) -> None:
        info(f"Fetching notebooks from local path `{repo_path}`")
        repo_path = Path(repo_path)
        if not repo_path.is_dir():
            raise OSError(f"`{repo_path}` is not a directory.")
        copy_docs_dirs(repo_path)

    notebooks_local_path = Path(
        os.environ.get(LIVECELLX_NOTEBOOK_PATH_ENV_VAR, CUR_DIR.absolute().parent.parent.parent / "notebooks")
    )
    try:
        fetch_local(notebooks_local_path)
    except Exception as e:
        warning(f"read`{notebooks_local_path}` failed, error message: `{e}`. Trying remote")
        require_download = int(os.environ.get(LIVECELLX_NOTEBOOK_PATH_ENV_VAR, 1))
        if not require_download:
            info(f"Used downloaded files as set in ENV")
            return

        fetch_remote(repo_url)



import logging
import numpy as np
import scipy.sparse as sp
from scipy.sparse import csr_matrix, issparse


def linear_embedding_velocity(V, PCs):
    return V @ PCs


def _estimate_dt(X, V, nbrs_idx):
    """Estimates the time step (`dt`) based on local density for each cell.

    This function calculates an estimate of the time step for each cell in the dataset by evaluating the 
    local density of neighboring cells. It assumes that the time step is inversely proportional to the 
    velocity magnitude relative to the local density of neighboring points.

    The process involves computing the Euclidean distance between each cell and its neighbors within local patch,
    then averaging these distances to estimate the local density. The time step is then estimated as the 
    median of the density-to-velocity ratio across neighbors for each cell.

    Parameters
    ----------
    X : :class:`~numpy.ndarray`
        The gene expression data matrix of shape `(n_cells, n_genes)`.
    V : :class:`~numpy.ndarray`
        The velocity matrix of shape `(n_cells, n_genes)`, containing the velocity vectors for each cell.
    nbrs_idx : list of list of int
        A list where each element contains the indices of the nearest neighbors for a given cell.
        Each cell's neighbors are identified using a nearest-neighbor search algorithm, typically based 
        on the gene expression data.

    Returns
    -------
    :class:`~numpy.ndarray`
        A column vector of shape `(n_cells, 1)` representing the estimated time steps (`dt`) for each cell.
    """
    logging.info('Estimate dt based on local density...')
    dt = np.zeros(X.shape[0])
    for i, ind in enumerate(nbrs_idx):
        delta_i = X[ind] - X[i]
        density = np.mean(np.linalg.norm(delta_i, axis=1))
        dt[i] = np.median(density / np.linalg.norm(V[i]))
    return dt.reshape(-1, 1)


def binary_corr(xi, Xj, vi):
    return np.mean(np.sign(vi) == np.sign(Xj - xi), axis=1)

def inner_product(xi, Xj, vi):
    D = Xj - xi
    return D @ vi 


def pearson_corr(xi, Xj, vi):
    D = Xj - xi
    Vi = np.tile(vi, (D.shape[0], 1))
    c = np.zeros(D.shape[0])
    for i in range(D.shape[0]):
        c[i] = np.corrcoef(D[i], Vi[i])[0, 1]

    return c


def cos_corr(xi, Xj, vi):
    D = Xj - xi
    dist = np.linalg.norm(D, axis=1)
    dist[dist == 0] = 1
    D /= dist[:, None]

    v_norm = np.linalg.norm(vi)
    if v_norm == 0:
        v_norm = 1
    
    return D @ vi / v_norm


def corr_kernel(X, V, nbrs, sigma=None, corr_func=cos_corr, softmax_adjusted=False):
    if softmax_adjusted:
        if sigma is None:
            logging.info("param sigma is `None`, estimate sigma automatically...")
            sigma = _estimate_sigma(X, V, nbrs, corr_func)
        logging.info(f"Using `sigma={sigma:.4f}`")
    P = np.zeros((X.shape[0], X.shape[0]))
    for i, x in enumerate(X):
        v, idx = V[i], nbrs[i]
        c = corr_func(x, X[idx], v)
        if softmax_adjusted:
            c = np.exp(c*sigma)
            c /= np.sum(c)
        P[i][idx] = c
    return P


def projection_with_transition_matrix(T, X_emb, correct_density=True, norm_diff=False):
    n = T.shape[0]
    V = np.zeros((n, X_emb.shape[1]))

    if not issparse(T):
        T = csr_matrix(T)

    for i in range(n):
        idx = T[i].indices
        diff_emb = X_emb[idx] - X_emb[i, None]
        if norm_diff:
            diff_emb /= np.linalg.norm(diff_emb, axis=1)[:, None]
        if np.isnan(diff_emb).sum() != 0:
            diff_emb[np.isnan(diff_emb)] = 0
        T_i = T[i].data
        V[i] = T_i.dot(diff_emb)
        if correct_density:
            V[i] -= T_i.mean() * diff_emb.sum(0)

    return V


def density_corrected_transition_matrix(T):
    T = sp.csr_matrix(T, copy=True)

    for i in range(T.shape[0]):
        idx = T[i].indices
        T_i = T[i].data
        T_i -= T_i.mean()
        T[i, idx] = T_i

    return T


def _estimate_sigma(X, V, nbrs, corr_func) -> float:
    """
    Adapted from cellrank
    """
    logits = np.zeros((X.shape[0], X.shape[0]))
    for i, x in enumerate(X):
        v, idx = V[i], nbrs[i]
        c = corr_func(x, X[idx], v)
        logits[i][idx] = c
    logits_abs = np.abs(logits)
    return 1.0 / np.median(logits_abs[logits_abs>0])


# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------
import os
import sys
from pathlib import Path
from datetime import datetime

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
from urllib.request import urlretrieve

module_path = os.path.join(os.path.dirname(__file__), "../..")
sys.path.insert(0, os.path.abspath(module_path))
sys.path.insert(0, os.path.abspath("../"))
sys.path.insert(0, os.path.abspath("../../"))
from source.docs_download_utils import _download_docs_dirs

# import graphvelo

# HERE = Path(__file__).parent
# sys.path[:0] = [str(HERE.parent)]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]
source_suffix = {".rst": "restructuredtext"}
# bibtex_bibfiles = ["./notebooks/lap.bib", "./notebooks/livetracker_ref.bib"]
# bibtex_reference_style = "author_year"

master_doc = "index"
pygments_style = "sphinx"


github_org = "xing-lab-pitt"
github_code_repo = "graphvelo"
github_ref = "master"
github_nb_repo = "graphvelo_notebooks"

# TODO: download and maintain notebooks from github
# _download_docs_dirs(repo_url=f"https://github.com/{github_org}/{github_nb_repo}")

# Add notebooks prolog to Google Colab and nbviewer
# TODO
nbsphinx_prolog = r"""
"""
nbsphinx_execute = "never"  # never execute notebooks

# -- Project information -----------------------------------------------------

project = 'GraphVelo'
package_name = "graphvelo"
author = "Chen, Yuhao and Zhang, Yan and Gan, Jiaqi and Ni, Ke and Chen, Ming and Bahar, Ivet and Xing, Jianhua"
copyright = f"{datetime.now():%Y}, {author}."
version = "0.1.0"
repository_url = "https://github.com/xing-lab-pitt/GraphVelo"

# The full version, including alpha/beta/rc tags
release = "0.1.0"

bibtex_bibfiles = ["references.bib"]
bibtex_reference_style = "author_year"
bibtex_default_style = "alpha"

templates_path = ["_templates"]
nitpicky = True  # Warn about broken links
needs_sphinx = "4.0"

html_context = {
    "display_github": True,  # Integrate GitHub
    "github_user": "xing-lab-pitt",  # Username
    "github_repo": project,  # Repo name
    "github_version": "main",  # Version
    "conf_py_path": "/docs/",  # Path in the checkout to the docs root
}
github_repo = "https://github.com/" + html_context["github_user"] + "/" + project


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.

extensions = [
    "nbsphinx",
    "sphinx.ext.autodoc",
    "sphinx.ext.napoleon",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.napoleon",
    # Link to other project's documentation (see mapping below)
    "sphinx.ext.intersphinx",
    # Add a link to the Python source code for classes, functions etc.
    "sphinx.ext.viewcode",
    "sphinx.ext.githubpages",
    "sphinx.ext.autosectionlabel",
    # Automatically document param types (less noise in class signature)
    "sphinx_autodoc_typehints",
    # "sphinxcontrib.bibtex",
    "sphinx_gallery.load_style",
]

# Mappings for sphinx.ext.intersphinx. Projects have to have Sphinx-generated doc! (.inv file)
intersphinx_mapping = {
    "anndata": ("https://anndata.readthedocs.io/en/stable/", None),
    "cycler": ("https://matplotlib.org/cycler/", None),
    "dask": ("https://docs.dask.org/en/latest/", None),
    "h5py": ("http://docs.h5py.org/en/stable/", None),
    "ipython": ("https://ipython.readthedocs.io/en/stable/", None),
    "louvain": ("https://louvain-igraph.readthedocs.io/en/latest/", None),
    "matplotlib": ("https://matplotlib.org/stable/", None),
    "networkx": (
        "https://networkx.org/documentation/stable/",
        None,
    ),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "pandas": ("https://pandas.pydata.org/pandas-docs/stable/", None),
    "pytest": ("https://docs.pytest.org/en/latest/", None),
    "python": ("https://docs.python.org/3", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/", None),
    "seaborn": ("https://seaborn.pydata.org/", None),
    "sklearn": ("https://scikit-learn.org/stable/", None),
}

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "**.ipynb_checkpoints"]

# extlinks config
extlinks = {
    "issue": (f"{repository_url}/issues/%s", "#%s"),
    "pr": (f"{repository_url}/pull/%s", "#%s"),
    "ghuser": ("https://github.com/%s", "@%s"),
}

# Generate the API documentation when building
autosummary_generate = True
autodoc_member_order = "bysource"
autoclass_content = "both"  # Add __init__ doc (ie. params) to class summaries
# Remove 'view source code' from top of page (for html, not python)
html_show_sourcelink = True
# If no class summary, inherit base class summary
autodoc_inherit_docstrings = True

autodoc_default_flags = [
    # Make sure that any autodoc declarations show the right members
    "members",
    "inherited-members",
    "private-members",
    "show-inheritance",
]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]
html_title = "GraphVelo"
# html_theme = "furo"
html_theme_options = dict(
    repository_url=github_repo,
    navigation_depth=4,
    use_repository_button=True,
    # logo_only=True,
)
html_context = dict(
    display_github=True,  # Integrate GitHub
    github_user=github_org,  # organization
    github_repo=github_code_repo,  # Repo name
    github_version="master",  # Version
    conf_py_path="/docs/source/",
)
html_show_sphinx = False
# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_logo = "_static/logo.png"


autodoc_member_order = "groupwise"
autodoc_typehints = "signature"
autodoc_docstring_signature = True
napoleon_google_docstring = False
napoleon_numpy_docstring = True
napoleon_include_init_with_doc = False
napoleon_use_rtype = True
napoleon_use_param = True


def setup(app):
    app.add_css_file("css/custom.css")


sphinx_enable_epub_build = False
sphinx_enable_pdf_build = False